#This script generates all the h5 example files
from pathlib import Path
from cadet import H5, Cadet
from addict import Dict
import numpy
import pandas
import CADETMatch.util

import create_example_sims

def create_experiments(defaults):
    dex_sim = create_example_sims.create_dextran_model(defaults)
    non_sim = create_example_sims.create_nonbinding_model(defaults)

    experiments_dir = defaults.base_dir / "experiments"

    single = experiments_dir / "single"
    single.mkdir(parents=True, exist_ok=True)

    multiple = experiments_dir / "multiple"
    multiple.mkdir(parents=True, exist_ok=True)

    advanced = experiments_dir / "multiple_advanced"
    advanced.mkdir(parents=True, exist_ok=True)

    dex_sim.filename = (single / "dextran.h5").as_posix()
    dex_sim.save()

    dex_sim.filename = (multiple / "dextran1.h5").as_posix()
    dex_sim.save()

    dex_sim.filename = (advanced / "dextran.h5").as_posix()
    dex_sim.save()

    print("Run ", dex_sim.run_load())

    times = dex_sim.root.output.solution.solution_times
    values = dex_sim.root.output.solution.unit_002.solution_outlet_comp_000
    data = numpy.array([times, values]).T

    numpy.savetxt(single / "dextran.csv", data, delimiter=',')
    numpy.savetxt(multiple / "dextran1.csv", data, delimiter=',')
    numpy.savetxt(advanced / "dextran.csv", data, delimiter=',')


    dex_sim.filename = (multiple / "dextran2.h5").as_posix()
    dex_sim.root.input.solver.sections.section_times = [0.0, 24.0, 600.0]
    dex_sim.root.input.model.unit_000.sec_000.const_coeff = [0.001,]
    dex_sim.save()

    print("Run ", dex_sim.run_load())

    times = dex_sim.root.output.solution.solution_times
    values = dex_sim.root.output.solution.unit_002.solution_outlet_comp_000
    data = numpy.array([times, values]).T

    numpy.savetxt(multiple / "dextran2.csv", data, delimiter=',')

    non_sim.filename = (advanced / "non.h5").as_posix()
    non_sim.save()

    print("Run ", non_sim.run_load())

    times = non_sim.root.output.solution.solution_times
    values = non_sim.root.output.solution.unit_002.solution_outlet_comp_000
    data = numpy.array([times, values]).T

    numpy.savetxt(advanced / "non.csv", data, delimiter=',')

def create_scores(defaults):
    #dextran scores
    dextran_paths = ['DextranShape', 'Shape', 'ShapeBack', 'ShapeFront', 'SSE', 
                     'other/curve', 'other/DextranSSE', 'other/ShapeDecay',
                     'other/ShapeDecayNoDer', 'other/ShapeDecaySimple', 'other/ShapeNoDer', 
                     'other/ShapeOnly', 'other/ShapeSimple', 'other/similarity', 
                     'other/similarityDecay', 'misc/multiple_scores_slicing']

    dex_sim = create_example_sims.create_dextran_model(defaults)

    scores_dir = defaults.base_dir / "scores"

    for path in dextran_paths:
        dir = scores_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        dex_sim.filename = (dir / "dextran.h5").as_posix()
        dex_sim.save()
        print("Run ", dir_name, dex_sim.run_load())

        times = dex_sim.root.output.solution.solution_times
        values = dex_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dextran.csv", data, delimiter=',')

    #flatten
    flat_sim = create_example_sims.create_dextran_model(defaults)
    flat_sim.root.input.model.unit_000.sec_000.const_coeff = [1.0,]
    flat_sim.root.input.model.unit_000.sec_001.const_coeff = [1.0,]
    dir = scores_dir / "Ceiling"
    dir.mkdir(parents=True, exist_ok=True)
    dir_name = dir.name
    flat_sim.filename = (dir / "flat.h5").as_posix()
    flat_sim.save()
    print("Run ", dir_name, flat_sim.run_load())

    times = flat_sim.root.output.solution.solution_times
    values = flat_sim.root.output.solution.unit_002.solution_outlet_comp_000
    data = numpy.array([times, values]).T

    numpy.savetxt(dir / "flat.csv", data, delimiter=',')


    #fractionate
    #use a 2-comp linear isotherm and densely sample the time to improve the estimation
    frac_sim = create_example_sims.create_linear_model(defaults)

    dirs = [scores_dir / "fractionationSlide", scores_dir / "other" / "fractionationSSE", scores_dir / "misc" / "multiple_components",
           defaults.base_dir / "transforms" / "misc" / "index"]

    for dir in dirs:
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        frac_sim.filename = (dir / "fraction.h5").as_posix()
        frac_sim.save()
        print("Run ", dir_name, frac_sim.run_load())

        times0 = frac_sim.root.output.solution.solution_times
        values0 = frac_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data0 = numpy.array([times0, values0]).T

        values1 = frac_sim.root.output.solution.unit_002.solution_outlet_comp_001
        data1 = numpy.array([times0, values1]).T

        data_sum = numpy.array([times0, values0 + values1]).T

        numpy.savetxt(dir / "comp0.csv", data0, delimiter=',')
        numpy.savetxt(dir / "comp1.csv", data1, delimiter=',')
        numpy.savetxt(dir / "data_sum.csv", data_sum, delimiter=',')

        times = numpy.linspace(400, 800, 9)
        start_times = times[:-1]
        stop_times = times[1:]
        fracs = CADETMatch.util.fractionate_sim(start_times, stop_times, [0,1], frac_sim, 'unit_002')

        df = pandas.DataFrame({'Start':start_times, 'Stop':stop_times, '0':fracs[0], '1':fracs[1]})
        df.to_csv((dir / "frac.csv").as_posix(), index=False, header=True)


def create_search(defaults):
    dex_sim = create_example_sims.create_dextran_model(defaults)

    dextran_paths = ['gradient', 'graphSpace', 'mcmc/stage1', 'multistart', 'nsga3', 'unsga3', 'scoretest',
                     'other/early_stopping', 'other/refine_shape', 'other/refine_sse']

    search_dir = defaults.base_dir / "search"

    for path in dextran_paths:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        dex_sim.filename = (dir / "dextran.h5").as_posix()
        dex_sim.save()
        print("Run ", dir_name, dex_sim.run_load())

        times = dex_sim.root.output.solution.solution_times
        values = dex_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dextran.csv", data, delimiter=',')


    non_sim = create_example_sims.create_nonbinding_model(defaults)

    non_paths = ['mcmc/stage2', ]

    search_dir = defaults.base_dir / "search"

    for path in non_paths:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        non_sim.filename = (dir / "non.h5").as_posix()
        non_sim.save()
        print("Run ", dir_name, non_sim.run_load())

        times = non_sim.root.output.solution.solution_times
        values = non_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "non.csv", data, delimiter=',')

def create_transforms(defaults):
    create_transforms_dextran(defaults)
    create_transforms_non(defaults)
    create_transforms_linear(defaults)
    create_transforms_sum(defaults)
    create_transforms_linear_exp(defaults)

def create_transforms_linear_exp(defaults):
    dex_sim1 = create_example_sims.create_dextran_model(defaults)

    dex_sim2 = create_example_sims.create_dextran_model(defaults)
    dex_sim2.root.input.model.unit_001.col_dispersion = 2 * dex_sim2.root.input.model.unit_001.col_dispersion

    dex_sim3 = create_example_sims.create_dextran_model(defaults)
    dex_sim3.root.input.model.unit_001.col_dispersion = 3 * dex_sim3.root.input.model.unit_001.col_dispersion

    search_dir = defaults.base_dir / "transforms"

    for path in ['norm_linear', 'other/linear']:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        
        dex_sim1.filename = (dir / "dex1.h5").as_posix()
        dex_sim1.save()
        print("Run ", dir_name, dex_sim1.run_load())
 
        times = dex_sim1.root.output.solution.solution_times
        values = dex_sim1.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dex1.csv", data, delimiter=',')

        dex_sim2.filename = (dir / "dex2.h5").as_posix()
        dex_sim2.save()
        print("Run ", dir_name, dex_sim2.run_load())

        times = dex_sim2.root.output.solution.solution_times
        values = dex_sim2.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dex2.csv", data, delimiter=',')

        dex_sim3.filename = (dir / "dex3.h5").as_posix()
        dex_sim3.save()
        print("Run ", dir_name, dex_sim3.run_load())

        times = dex_sim3.root.output.solution.solution_times
        values = dex_sim3.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dex3.csv", data, delimiter=',')



def create_transforms_sum(defaults):
    cstr_sim = create_example_sims.create_cstr_model(defaults)

    search_dir = defaults.base_dir / "transforms"

    dir = search_dir / "sum"
    dir.mkdir(parents=True, exist_ok=True)
    dir_name = dir.name
    cstr_sim.filename = (dir / "cstr.h5").as_posix()
    cstr_sim.save()
    print("Run ", dir_name, cstr_sim.run_load())

    times = cstr_sim.root.output.solution.solution_times
    values = cstr_sim.root.output.solution.unit_002.solution_outlet_comp_000
    data = numpy.array([times, values]).T

    numpy.savetxt(dir / "cstr.csv", data, delimiter=',')


def create_transforms_linear(defaults):
    lin_sim = create_example_sims.create_linear_model(defaults)

    lin_paths = ['auto_keq', 'other/keq', 'other/norm_keq', 'set_value']

    search_dir = defaults.base_dir / "transforms"

    for path in lin_paths:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        lin_sim.filename = (dir / "lin.h5").as_posix()
        lin_sim.save()
        print("Run ", dir_name, lin_sim.run_load())

        times = lin_sim.root.output.solution.solution_times
        values = lin_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "lin.csv", data, delimiter=',')

def create_transforms_non(defaults):
    non_sim = create_example_sims.create_nonbinding_model(defaults)

    non_paths = ['auto_inverse', 'norm_add', 'norm_mult', ]

    search_dir = defaults.base_dir / "transforms"

    for path in non_paths:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        non_sim.filename = (dir / "non.h5").as_posix()
        non_sim.save()
        print("Run ", dir_name, non_sim.run_load())

        times = non_sim.root.output.solution.solution_times
        values = non_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "non.csv", data, delimiter=',')
    
def create_transforms_dextran(defaults):
    dex_sim = create_example_sims.create_dextran_model(defaults)

    dextran_paths = ['auto', 'norm_diameter', 'norm_volume_area', 'norm_volume_length', 
                     'other/diameter', 'other/log', 'other/norm', 'other/norm_log', 'other/null', 
                     'other/volume_area', 'other/volume_length']

    search_dir = defaults.base_dir / "transforms"

    for path in dextran_paths:
        dir = search_dir / path
        dir.mkdir(parents=True, exist_ok=True)
        dir_name = dir.name
        dex_sim.filename = (dir / "dextran.h5").as_posix()
        dex_sim.save()
        print("Run ", dir_name, dex_sim.run_load())

        times = dex_sim.root.output.solution.solution_times
        values = dex_sim.root.output.solution.unit_002.solution_outlet_comp_000
        data = numpy.array([times, values]).T

        numpy.savetxt(dir / "dextran.csv", data, delimiter=',')

def main(defaults):
    "create simulations by directory"
    create_experiments(defaults)
    create_scores(defaults)
    create_search(defaults)
    create_transforms(defaults)

if __name__ == "__main__":
    main()



import csv
import logging

# parallelization
import multiprocessing
import shutil
import sys
import time
from pathlib import Path

import h5py
import numpy
import functools
from cadet import H5, Cadet

import CADETMatch.evo as evo
import CADETMatch.gradFD as gradFD
import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.util as util
import CADETMatch.version as version
from CADETMatch.cache import cache


def main(map_function):
    path = sys.argv[1]
    setup(cache, path, map_function)
    gradFD.setupTemplates(cache)
    hof = evo.run(cache)

    multiprocessing.get_logger().info("altScores %s", cache.altScores)
    multiprocessing.get_logger().info("altScoreNames %s", cache.altScoreNames)

    if cache.altScores:
        for name in cache.altScoreNames:
            json_path = util.setupAltFeature(cache, name)
            setup(cache, json_path, map_function)
            hof = evo.run(cache)

    continue_mcmc(cache, map_function)

    if "repeat" in cache.settings:
        repeat = int(cache.settings["repeat"])

        for i in range(repeat):
            json_path = util.repeatSimulation(i)
            multiprocessing.get_logger().info(json_path)

            setup(cache, json_path)

            hof = evo.run(cache)

        util.metaCSV(cache)

    if "bootstrap" in cache.settings:
        temp = []

        samples = int(cache.settings["bootstrap"]["samples"])

        if samples:

            center = float(cache.settings["bootstrap"]["center"])
            noise = float(cache.settings["bootstrap"]["percentNoise"]) / 100.0

            bootstrap = cache.settings["resultsDirBase"] / "bootstrap_output"

            for i in range(samples):
                # copy csv files to a new directory with noise added
                # put a new json file in the directory that points to the new csv files
                json_path = util.copyCSVWithNoise(i, center, noise)
                multiprocessing.get_logger().info(json_path)

                setup(cache, json_path)

                # call setup on all processes with the new json file as an argument to reset them
                # util.updateScores(json_path)

                hof = evo.run(cache)
                temp.append(util.bestMinScore(hof))

                numpy_temp = numpy.array(temp)
                cov = numpy.cov(numpy_temp.transpose())
                multiprocessing.get_logger().info(
                    "in progress cov %s data %s det %s",
                    cov,
                    numpy_temp,
                    numpy.linalg.det(cov),
                )

            numpy_temp = numpy.array(temp)
            cov = numpy.cov(numpy_temp.transpose())
            multiprocessing.get_logger().info(
                "final cov %s data %s det %s", cov, numpy_temp, numpy.linalg.det(cov)
            )


def setup(cache, json_path, map_function):
    "run seutp for the current json_file"
    cache.setup_dir(json_path)

    createDirectories(cache, json_path)
    util.setupLog(cache.settings["resultsDirLog"], "main.log")

    print_version()

    cache.setup(json_path)
    cache.map_function = map_function

    cache.eval.evaluate = functools.partial(evo.fitness, json_path=json_path)
    cache.eval.evaluate_final = functools.partial(evo.fitness_final, json_path=json_path)
    cache.eval.evaluate_grad = functools.partial(gradFD.gradSearch, json_path=json_path)
    cache.eval.evaluate_grad_fine = functools.partial(gradFD.gradSearchFine, json_path=json_path)
    cache.eval.grad_search = gradFD.search

    createCSV(cache)
    createProgressCSV(cache)
    createErrorCSV(cache)
    setupTemplates(cache)


def print_version():
    multiprocessing.get_logger().info(
        "CADETMatch starting up version: %s", version.__version__
    )

    import importlib_metadata

    modules = [
        ('attrs', '21.2.0'),
        ("joblib", "1.0.1"),
        ("addict", "2.4.0"),
        ("corner", "2.2.1"),
        ("emcee", "3.0.2"),
        ("SALib", "1.3.11"),
        ("psutil", "5.8.0"),
        ("numpy", "1.21.1"),
        ("openpyxl", "3.0.7"),
        ("scipy", "1.7.0"),
        ("matplotlib", "3.4.2"),
        ("pandas", "1.3.0"),
        ("h5py", "3.3.0"),
        ("cadet-python", "0.11"),
        ("seaborn", "0.11.1"),
        ("scikit-learn", "0.24.2"),
        ("jstyleson", "0.2.0"),
        ('filelock', "3.0.12"),
        ('pymoo', '0.4.2.2')
    ]

    for module, version_tested in modules:
        multiprocessing.get_logger().info(
            "%s version: %s tested with %s",
            module,
            importlib_metadata.version(module),
            version_tested,
        )


def createDirectories(cache, json_path):
    cache.settings["resultsDirBase"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirGrad"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirMisc"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirSpace"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirEvo"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirProgress"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirMeta"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirLog"].mkdir(parents=True, exist_ok=True)
    cache.settings["resultsDirMCMC"].mkdir(parents=True, exist_ok=True)

    # copy simulation setting file to result base directory
    try:
        shutil.copy(str(json_path), str(cache.settings["resultsDirBase"]))
    except shutil.SameFileError:
        pass


def createCSV(cache):
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    if not path.exists():
        with path.open("w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
            writer.writerow(cache.headers)

    path = cache.settings["resultsDirMeta"] / "results.csv"
    if not path.exists():
        with path.open("w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
            writer.writerow(cache.headers)


def createProgressCSV(cache):
    path = Path(cache.settings["resultsDirBase"], "progress.csv")
    cache.progress_path = path
    if not path.exists():
        with path.open("w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
            writer.writerow(cache.progress_headers)


def createErrorCSV(cache):
    path = Path(cache.settings["resultsDirBase"], "error.csv")
    cache.error_path = path
    if not path.exists():
        with path.open("w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
            writer.writerow(
                cache.parameter_headers + ["Return Code", "STDOUT", "STDERR"]
            )


def setTemplateValues(simulation, set_values):
    for path, index, value in set_values:
        if index >= 0:
            simulation[path][index] = value
        else:
            simulation[path] = value


def setTemplateValuesAuto(simulation, set_values_auto, cache):
    if "mcmc_h5" not in cache.settings:
        multiprocessing.get_logger().error(
            "set_values_auto can't be used without mcmc_h5 as a prior"
        )

    mcmc_h5 = Path(cache.settings["mcmc_h5"])
    mle_h5 = mcmc_h5.parent / "mle.h5"

    data = H5()
    data.filename = mle_h5.as_posix()
    data.load(lock=True)
    stat_MLE = data.root.stat_MLE

    used = set()

    for path, index, mle_index in set_values_auto:
        value = stat_MLE[mle_index]
        if mle_index not in used:
            used.add(mle_index)
        if index >= 0:
            simulation[path][index] = value
        else:
            simulation[path] = value

    if len(used) != len(stat_MLE):
        multiprocessing.get_logger().warn(
            "not all values from the prior where used, proceed with caution"
        )


def setupTemplates(cache):
    "setup all the experimental templates"
    for experiment in cache.settings["experiments"]:
        HDF5 = experiment["HDF5"]
        name = experiment["name"]

        template_path = Path(cache.settings["resultsDirMisc"], "template_%s.h5" % name)

        template = Cadet()

        # load based on where the HDF5 file is
        template.filename = HDF5
        template.load()
        template.root.experiment_name = name

        if "set_values" in experiment:
            setTemplateValues(template, experiment["set_values"])

        if "set_values_auto" in experiment:
            setTemplateValuesAuto(template, experiment["set_values_auto"], cache)

        util.setupSimulation(template, cache.target[name]["time"], name, cache)

        start = time.time()
        util.runExperiment(
            None,
            experiment,
            cache.settings,
            cache.target,
            template,
            experiment.get("timeout", 1800),
            cache,
        )
        elapsed = time.time() - start

        multiprocessing.get_logger().info("simulation took %s", elapsed)

        # timeout needs to be stored in the template so all processes have it without calculating it
        template.root.timeout = max(10, elapsed * 10)

        if (
            cache.settings["searchMethod"] != "MCMC"
            and "errorModel" in cache.settings
        ):
            # the base case needs to be saved since the normal template file is what the rest of the code will look for
            template_base_path = Path(
                cache.settings["resultsDirMisc"], "template_%s_base.h5" % name
            )
            template.filename = template_base_path.as_posix()
            template.save()

            multiprocessing.get_logger().info(
                "create bias template for experiment %s", name
            )
            template_bias = util.biasSimulation(template, experiment, cache)
            template_bias_path = Path(
                cache.settings["resultsDirMisc"], "template_%s_bias.h5" % name
            )
            template_bias.filename = template_bias_path.as_posix()
            template_bias.save()
            template = template_bias

        # change to where we want the template created
        template.filename = template_path.as_posix()

        template.save()

        experiment["simulation"] = template

        template_path = Path(
            cache.settings["resultsDirMisc"], "template_%s_final.h5" % name
        )
        template_final = Cadet(template.root)
        template_final.filename = template_path.as_posix()
        if cache.dynamicTolerance:
            template_final.root.input.solver.time_integrator.abstol = (
                util.get_grad_tolerance(cache, name)
            )
            template_final.root.input.solver.time_integrator.reltol = 0.0

        start = time.time()
        util.runExperiment(
            None,
            experiment,
            cache.settings,
            cache.target,
            template_final,
            experiment.get("timeout", 1800),
            cache,
        )
        elapsed = time.time() - start

        multiprocessing.get_logger().info("simulation final took %s", elapsed)

        # timeout needs to be stored in the template so all processes have it without calculating it
        template_final.root.timeout = max(10, elapsed * 10)

        template_final.save()
        experiment["simulation_final"] = template_final


def continue_mcmc(cache, map_function):
    if cache.continueMCMC:
        json_path = util.setupMCMC(cache)
        multiprocessing.get_logger().info(json_path)

        setup(cache, json_path, map_function)

        hof = evo.run(cache)


if __name__ == "__main__":
    start = time.time()
    map_function = util.getMapFunction()
    main(map_function=map_function)
    multiprocessing.get_logger().info("System has finished")
    multiprocessing.get_logger().info(
        "The total runtime was %s seconds" % (time.time() - start)
    )
    sys.exit()



import array
import csv
import multiprocessing
import random
import time
from pathlib import Path

import CADETMatch.pareto as pareto
import CADETMatch.progress as progress
import CADETMatch.sub as sub
import CADETMatch.util as util
import CADETMatch.pop as pop
import numpy

name = "Multistart"


def run(cache):
    "run the parameter estimation"
    random.seed()

    parameters = len(cache.MIN_VALUE)

    populationSize = parameters * cache.settings["population"]
    sim_start = generation_start = time.time()
    result_data = {
        "input": [],
        "output": [],
        "output_meta": [],
        "results": {},
        "times": {},
        "input_transform": [],
        "input_transform_extended": [],
        "strategy": [],
        "mean": [],
        "confidence": [],
    }

    init_pop = util.sobolPopulation(populationSize, parameters, numpy.array(cache.MIN_VALUE), numpy.array(cache.MAX_VALUE))
    init_pop = [pop.Individual(row) for row in init_pop]

    if "seeds" in cache.settings:
        seed_pop = [
            pop.Individual(
                [f(v) for f, v in zip(cache.settings["transform"], sublist)]
            )
            for sublist in cache.settings["seeds"]
        ]
        init_pop.extend(seed_pop)

    gradCheck = cache.badScore

    if cache.metaResultsOnly:
        hof = pareto.DummyFront()
    else:
        hof = pareto.ParetoFront(dimensions=len(cache.WORST),
            similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
        )
    meta_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
        similar=pareto.similar,
        similar_fit=pareto.similar_fit_meta(cache),
        slice_object=cache.meta_slice,
    )
    grad_hof = pareto.ParetoFront(dimensions=len(cache.WORST),
        similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
    )
    progress_hof = pareto.DummyFront()

    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        multiprocessing.get_logger().info(
            "Population %s", [util.convert_individual_inputorder(i, cache) for i in init_pop]
        )

        gradCheck, newChildren = cache.eval.grad_search(
            gradCheck, init_pop, cache, writer, csvfile, hof, meta_hof, -1, check_all=True
        )

        stalled, stallWarn, progressWarn = util.eval_population(
            cache,
            newChildren,
            writer,
            csvfile,
            hof,
            meta_hof,
            None,
            -1,
            result_data,
        )

        progress.writeProgress(
            cache,
            -1,
            newChildren,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        util.finish(cache)
        sub.graph_corner_process(cache, last=True)

        return hof



import sys

import matplotlib
import matplotlib.style as mplstyle

mplstyle.use("fast")

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

import itertools
import logging

# parallelization
import multiprocessing
import os
import warnings
from pathlib import Path

import matplotlib.cm
import numpy
import scipy.interpolate
from addict import Dict
from cadet import H5, Cadet
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.colors import ListedColormap
from mpl_toolkits.mplot3d import Axes3D
from sklearn.preprocessing import MinMaxScaler

import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.util as util
from CADETMatch.cache import cache

cm_plot = matplotlib.cm.gist_rainbow


def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "mixing.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("mixing graphing directory %s", os.getcwd())

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    if mcmc_h5.exists():
        mcmc_store = H5()
        mcmc_store.filename = mcmc_h5.as_posix()
        mcmc_store.load(
            paths=[
                "/full_chain",
                "/train_full_chain",
                "/bounds_full_chain",
                "/full_chain_transform",
                "/train_full_chain_transform",
                "/bounds_full_chain_transform",
                "/bounds_probability",
                "/train_probability",
                "/run_probability",
            ],
            lock=True
        )

        progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"

        graph_dir = cache.settings["resultsDirSpace"] / "mcmc"

        mixing = graph_dir / "mixing"
        mixing.mkdir(parents=True, exist_ok=True)

        input_headers = cache.parameter_headers_actual

        chain_names = (
            ("full_chain", "run_probability"),
            ("train_full_chain", "train_probability"),
            ("bounds_full_chain", "bounds_probability"),
            ("full_chain_transform", "run_probability"),
            ("train_full_chain_transform", "train_probability"),
            ("bounds_full_chain_transform", "bounds_probability"),
        )

        for chain, probability in chain_names:
            if chain in mcmc_store.root:
                plot_chain(
                    input_headers,
                    mcmc_store.root[chain],
                    mcmc_store.root[probability],
                    mixing / ("mixing_%s" % chain),
                )


def plot_chain(headers, chain, probability, graph_dir):
    graph_dir.mkdir(parents=True, exist_ok=True)

    for i in range(chain.shape[2]):
        graph_mix(headers[i], chain[:, :, i], graph_dir)
        graph_probability(headers[i], chain[:, :, i], probability, graph_dir)


def graph_mix(header, chain, graph_dir):
    fig = figure.Figure(figsize=[15, 7])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    chain_length = chain.shape[1]
    x = numpy.linspace(0, chain_length - 1, chain_length)

    lines = []
    for j in range(chain.shape[0]):
        graph.plot(x, chain[j, :], color=get_color(j, chain.shape[0] - 1, cm_plot))

    graph.set_xlabel("chain length")
    graph.set_ylabel("value")
    filename = "mixing_%s.png" % (header)
    filename = filename.replace(":", "_").replace("/", "_")

    graph.set_title("Mixing graph %s" % header)
    fig.savefig((graph_dir / filename).as_posix())


def graph_probability(header, chain, probability, graph_dir):
    scaler = MinMaxScaler()
    probability = scaler.fit_transform(probability)

    fig = figure.Figure(figsize=[15, 7])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    chain_length = chain.shape[1]
    x = numpy.linspace(0, chain_length - 1, chain_length)

    lines = []
    for j in range(chain.shape[0]):
        graph.scatter(x, chain[j, :], color=matplotlib.cm.rainbow(probability[j, :]))

    graph.set_xlabel("chain length")
    graph.set_ylabel("value")
    filename = "mixing_%s_probability.png" % (header)
    filename = filename.replace(":", "_").replace("/", "_")

    graph.set_title("Mixing graph %s" % header)
    fig.savefig((graph_dir / filename).as_posix())


if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import csv
import time
import numpy
import random
import pickle
from pathlib import Path

import CADETMatch.pareto as pareto
import CADETMatch.progress as progress
import CADETMatch.sub as sub
import CADETMatch.util as util
import CADETMatch.pop as pop

from pymoo.factory import get_algorithm, get_reference_directions
from pymoo.core.problem import Problem
import attr


@attr.s
class ProblemState:
    cache = attr.ib()
    writer = attr.ib()
    csvfile = attr.ib()
    halloffame = attr.ib()
    meta_hof = attr.ib()
    progress_hof = attr.ib()
    result_data = attr.ib()


name = "pymoo"

class MyProblem(Problem):

    def __init__(self, n_var, n_obj, lb, ub, evaluate_function, problem_state):
        super().__init__(n_var=n_var, n_obj=n_obj, n_constr=0, xl=lb, xu=ub)
        self.evaluate_function = evaluate_function
        self.problem_state = problem_state

    def _evaluate(self, population, out, *args, **kwargs):
        population = [pop.Individual(row) for row in numpy.array(population)]

        stalled, stallWarn, progressWarn = util.eval_population(
                    self.problem_state.cache,
                    population,
                    self.problem_state.writer,
                    self.problem_state.csvfile,
                    self.problem_state.halloffame,
                    self.problem_state.meta_hof,
                    self.problem_state.progress_hof,
                    kwargs['algorithm'].n_gen,
                    self.problem_state.result_data,
                )

        self.stalled = stalled
        self.stallWarn = stallWarn
        self.progressWarn = progressWarn
        self.eval_population = population
        self.ngen = kwargs['algorithm'].n_gen

        error = numpy.array([ind.fitness.values for ind in population])
        out["F"] = error


def stop_iteration(best, stalled, cache):
    stopAverage = cache.settings.get("stopAverage", 0.0)
    stopBest = cache.settings.get("stopBest", 0.0)
    stopRMSE = cache.settings.get("stopRMSE", 0.0)
    if best[2] <= stopAverage or best[1] <= stopBest or stalled or best[-1] <= stopRMSE:
        return True
    else:
        return False


def run(cache, alg="unsga3"):
    "run the parameter estimation"
    random.seed()
    parameters = len(cache.MIN_VALUE)

    populationSize = parameters * cache.settings.get("population", 100)

    totalGenerations = parameters * cache.settings.get("generations", 1000)

    init_pop = util.sobolPopulation(populationSize, parameters, numpy.array(cache.MIN_VALUE), numpy.array(cache.MAX_VALUE))

    if "seeds" in cache.settings:
        seed_pop = [
            [f(v) for f, v in zip(cache.settings["transform"], sublist)]
            for sublist in cache.settings["seeds"]
        ]
        init_pop = numpy.concatenate([init_pop, numpy.array(seed_pop)], axis=0)

    
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        checkpointFile = Path(
            cache.settings["resultsDirMisc"], cache.settings.get("checkpointFile", "check")
        )

        pymoo_checkpointFile = Path(
            cache.settings["resultsDirMisc"], "pymoo_check.npy"
        )

        if checkpointFile.exists():
            with checkpointFile.open("rb") as cp_file:
                cp = pickle.load(cp_file)
            hof = cp["halloffame"]
            meta_hof = cp["meta_halloffame"]
            grad_hof = cp["grad_halloffame"]
            progress_hof = cp["progress_halloffame"]
            cache.generationsOfProgress = cp["generationsOfProgress"]
            cache.lastProgressGeneration = cp["lastProgressGeneration"]

            if cp["gradCheck"] > cache.settings.get("gradCheck", 0.0):
                gradCheck = cp["gradCheck"]
            else:
                gradCheck = cache.settings.get("gradCheck", 0.0)

        else:
            if cache.metaResultsOnly:
                hof = pareto.DummyFront()
            else:
                hof = pareto.ParetoFront(dimensions=len(cache.WORST),
                    similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
                )
            meta_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
                slice_object=cache.meta_slice
            )
            grad_hof = pareto.ParetoFront(dimensions=len(cache.WORST),
                similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
            )
            progress_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
                similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
                slice_object=cache.meta_slice
            )

            gradCheck = cache.settings.get("gradCheck", 0.0)

        sim_start = generation_start = time.time()
        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
        }
        
        problem_state = ProblemState(cache=cache, writer=writer, csvfile=csvfile, halloffame=hof,
                                     meta_hof=meta_hof, progress_hof=progress_hof, result_data=result_data)

        problem = MyProblem(parameters, cache.numGoals, cache.MIN_VALUE, cache.MAX_VALUE, cache.eval.evaluate, problem_state)

        numRefDirs = min(populationSize, 1000)

        ref_dirs = get_reference_directions("energy", cache.numGoals, numRefDirs, seed=1)


        if pymoo_checkpointFile.exists():
            algorithm, = numpy.load(pymoo_checkpointFile, allow_pickle=True).flatten()
            algorithm.problem = problem

            algorithm.n_offsprings = populationSize
            algorithm.pop_size = populationSize
        else:
            algorithm = get_algorithm(alg, ref_dirs=ref_dirs, sampling=init_pop, pop_size=populationSize )
            algorithm.setup(problem, termination=('n_gen', totalGenerations), seed=1)

        while algorithm.has_next():
            algorithm.next()

            stalled = problem.stalled
            stallWarn = problem.stallWarn
            progressWarn = problem.progressWarn
            population = problem.eval_population

            progress.writeProgress(
                cache,
                problem.ngen,
                population,
                hof,
                meta_hof,
                grad_hof,
                progress_hof,
                sim_start,
                generation_start,
                problem.problem_state.result_data,
            )

            sub.graph_process(cache, problem.ngen)
            sub.graph_corner_process(cache, last=False)

            best = meta_hof.getBestScores()
            cp = dict(
                population=population,
                halloffame=hof,
                rndstate=random.getstate(),
                meta_halloffame=meta_hof,
                grad_halloffame=grad_hof,
                gradCheck=gradCheck,
                generationsOfProgress=cache.generationsOfProgress,
                lastProgressGeneration=cache.lastProgressGeneration,
                progress_halloffame=progress_hof,
            )

            with checkpointFile.open("wb") as cp_file:
                pickle.dump(cp, cp_file)

            #can't pickle the pool and other things so all of that is just stored in problem_state
            #just temporarily unhook it from the problem and then add it back
            del algorithm.problem.problem_state
            numpy.save(pymoo_checkpointFile, algorithm)
            algorithm.problem.problem_state = problem_state

            if stop_iteration(best, stalled, cache):
                break

        gen = problem.ngen + 1
        if cache.finalGradRefinement:            
            best_individuals = [pop.Individual(i) for i in meta_hof]
            gradCheck, newChildren = cache.eval.grad_search(
                gradCheck,
                best_individuals,
                cache,
                writer,
                csvfile,
                grad_hof,
                meta_hof,
                gen,
                check_all=True,
                result_data=result_data,
            )
            if newChildren:
                progress.writeProgress(
                    cache,
                    gen,
                    newChildren,
                    hof,
                    meta_hof,
                    grad_hof,
                    progress_hof,
                    sim_start,
                    generation_start,
                    result_data,
                )

        population = [pop.Individual(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache,
            population,
            writer,
            csvfile,
            hof,
            meta_hof,
            progress_hof,
            gen + 1,
            result_data,
        )
        progress.writeProgress(
            cache,
            gen + 1,
            population,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        util.finish(cache)
        sub.graph_corner_process(cache, last=True)
        return hof



import copy
import csv
import decimal as decim
import hashlib
import multiprocessing
import os
import random
import subprocess
import sys
import tempfile
import time
import warnings
from pathlib import Path

import jstyleson
import numpy
import pandas
import psutil
import scipy.signal
from addict import Dict
from cadet import H5, Cadet

import CADETMatch.calc_coeff as calc_coeff
import CADETMatch.sub as sub
import CADETMatch.pop as pop

decim.getcontext().prec = 64
__logBase10of2_decim = decim.Decimal(2).log10()
__logBase10of2 = float(__logBase10of2_decim)

import logging
import os

import SALib.sample.sobol_sequence
import scipy.interpolate

import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.pareto as pareto
import CADETMatch.synthetic_error as synthetic_error

# numpy.warnings.filterwarnings('error', category=numpy.VisibleDeprecationWarning)


def get_times_values(simulation, target, selected=None):

    try:
        times = simulation.root.output.solution.solution_times

        output_path = target["output_path"]

        if isinstance(output_path, list):
            values = numpy.sum([simulation[i] for i in output_path], 0)
        else:
            values = simulation[output_path]
    except (AttributeError, KeyError):
        times = simulation[:, 0]
        values = simulation[:, 1]

    if selected is None:
        selected = target["selected"]
    return times[selected], values[selected] * target["factor"]


def sse(data1, data2):
    return numpy.sum((data1 - data2) ** 2)


def find_peak(times, data):
    "Return tuples of (times,data) for the peak we need"
    minIdx = numpy.argmin(data)
    maxIdx = numpy.argmax(data)

    return (times[maxIdx], data[maxIdx]), (times[minIdx], data[minIdx])


def find_breakthrough(times, data):
    "return tuple of time,value for the start breakthrough and end breakthrough"
    selected = data > 0.999 * max(data)
    selected_times = times[selected]
    return (selected_times[0], max(data)), (selected_times[-1], max(data))


def generateIndividual(icls, size, imin, imax, cache):
    return icls(numpy.random.uniform(imin, imax))


def initIndividual(icls, cache, content):
    return icls(content)

def sobolPopulation(populationSize, populationDimension, lb, ub):
    sobol = SALib.sample.sobol_sequence.sample(populationSize, populationDimension)
    sobol = sobol * (ub - lb) + lb
    return sobol


def sobolGenerator(icls, cache, n):
    if n > 0:
        populationDimension = len(cache.MIN_VALUE)
        populationSize = n
        sobol = sobolPopulation(populationSize, populationDimension, numpy.array(cache.MIN_VALUE), numpy.array(cache.MAX_VALUE))
        data = numpy.apply_along_axis(list, 1, sobol)
        data = list(map(icls, data))
        return data
    else:
        return []


def calcMetaScores(scores, cache):
    scores = numpy.array(scores)[cache.meta_mask]
    if cache.allScoreSSE and not cache.MultiObjectiveSSE:
        sse = numpy.sum(numpy.array(scores))
        return [sse, sse, sse]
    else:
        prod_score = product_score(scores)
        min_score = max(scores)
        mean_score = sum(scores) / len(scores)
        return [prod_score, min_score, mean_score]


def product_score(values):
    values = numpy.array(values)
    return 1.0 - scipy.stats.gmean(1-values)


def saveExperiments(save_name_base, settings, target, results, directory, file_pattern):
    for experiment in settings["experiments"]:
        experimentName = experiment["name"]
        simulation = results[experimentName]["simulation"]
        try:
            #for addict 2.4.0 otherwise when variables are set below it can't find __frozen even though it is not used
            simulation.root.unfreeze()
        except TypeError:
            pass

        dst = Path(directory, file_pattern % (save_name_base, experimentName))

        same = False
        if dst.is_file():
            # load tolerance to see if the files are actually the same
            sim_dst = Cadet()
            sim_dst.filename = dst.as_posix()
            sim_dst.load(
                paths=[
                    "/input/solver/time_integrator",
                ]
            )

            dst_tol = sim_dst.root.input.solver.time_integrator.abstol
            sim_tol = simulation.root.input.solver.time_integrator.abstol
            same = sim_tol == dst_tol

        if same:
            return False
        else:
            simulation.filename = dst.as_posix()

            for (header, score) in zip(
                experiment["headers"], results[experimentName]["scores"]
            ):
                simulation.root.score[header] = score
            simulation.save()

    return True


def convert_individual_inputorder(individual, cache):
    cadetValues = []

    idx = 0
    for parameter in cache.parameters:
        count = parameter.count
        if count:
            seq = individual[idx : idx + count]
            values = parameter.untransform_inputorder(seq)
            cadetValues.extend(values)
            idx += count

    return cadetValues


def convert_individual(individual, cache):
    cadetValues = []
    cadetValuesExtended = []

    idx = 0
    for parameter in cache.parameters:
        count = parameter.count
        if count:
            seq = individual[idx : idx + count]
            values, headerValues = parameter.untransform(seq)
            cadetValues.extend(values)
            cadetValuesExtended.extend(headerValues)
            idx += count

    return cadetValues, cadetValuesExtended


def convert_individual_grad(individual, cache):
    cadetValues = []

    idx = 0
    for parameter in cache.parameters:
        count = parameter.count
        if count:
            seq = individual[idx : idx + count]
            values = parameter.grad_untransform(seq)
            cadetValues.extend(values)
            idx += count

    return cadetValues


def convert_population(population, cache):
    cadetValues = numpy.zeros(population.shape)

    idx = 0
    for parameter in cache.parameters:
        count = parameter.count
        if count:
            matrix = population[:, idx : idx + count]
            values = parameter.untransform_matrix(matrix)
            cadetValues[:, idx : idx + count] = values
            idx += count

    return cadetValues


def convert_population_inputorder(population, cache):
    if len(population):
        cadetValues = numpy.zeros(population.shape)

        idx = 0
        for parameter in cache.parameters:
            count = parameter.count
            if count:
                matrix = population[:, idx : idx + count]
                values = parameter.untransform_matrix_inputorder(matrix)
                cadetValues[:, idx : idx + count] = values
                idx += count

        return cadetValues
    else:
        return numpy.array([])


def convert_individual_inverse(individual, cache):
    return numpy.array([f(v) for f, v in zip(cache.settings["transform"], individual)])


def convert_individual_inverse_grad(individual, cache):
    return numpy.array(
        [f(v) for f, v in zip(cache.settings["grad_transform"], individual)]
    )


def set_simulation(individual, simulation, settings, cache, experiment):
    multiprocessing.get_logger().debug("individual %s", individual)

    cadetValues = []
    cadetValuesKEQ = []

    idx = 0
    for parameter in cache.parameters:
        count = parameter.count
        # even if count is 0 this needs to be run so that setSimulation will be run for things like calculations
        seq = individual[idx : idx + count]
        values, headerValues = parameter.setSimulation(simulation, seq, experiment)
        cadetValues.extend(values)
        cadetValuesKEQ.extend(headerValues)
        idx += count

    multiprocessing.get_logger().debug("finished setting hdf5")
    return cadetValues, cadetValuesKEQ


def getBoundOffset(unit):
    if unit.unit_type == b"CSTR":
        NBOUND = unit.nbound
    else:
        NBOUND = unit.discretization.nbound

    if not len(NBOUND):
        "If NBOUND is empty it is all zero"
        NBOUND = [0] * unit.ncomp

    boundOffset = numpy.cumsum(
        numpy.concatenate(
            [
                [
                    0,
                ],
                NBOUND,
            ]
        )
    )
    return boundOffset


def runExperiment(
    individual,
    experiment,
    settings,
    target,
    template_sim,
    timeout,
    cache,
    post_function=None,
):
    simulation = Cadet(template_sim.root)

    if simulation.is_file:
        handle, path = tempfile.mkstemp(suffix=".h5", dir=cache.tempDir)
        os.close(handle)
    
        simulation.filename = path

    simulation.root.input.solver.nthreads = int(settings.get("nThreads", 1))

    if individual is not None:
        cadetValues, cadetValuesKEQ = set_simulation(
            individual, simulation, settings, cache, experiment
        )
    else:
        cadetValues = []
        cadetValuesKEQ = []

    if simulation.is_file:
        simulation.save()

    try:
        simulation.run(timeout=timeout, check=True)
    except subprocess.TimeoutExpired:
        multiprocessing.get_logger().warn("Simulation Timed Out")
        simulation.clear()
        if simulation.is_file:
            os.remove(path)
        return None

    except subprocess.CalledProcessError as error:
        multiprocessing.get_logger().error("The simulation failed %s", individual)
        logError(cache, cadetValuesKEQ, error)
        simulation.clear()
        return None

    # read sim data
    simulation.load_results()
    simulation.clear()

    if simulation.is_file:
        os.remove(path)

    user_solution_times = simulation.root.input.solver.user_solution_times
    simulationFailed = len(simulation.root.output.solution.solution_times) != len(user_solution_times)
    if simulationFailed:
        if not simulation.is_file:
            path = "dll interface"
        multiprocessing.get_logger().error(
            "%s sim must have failed %s", individual, path
        )
        return None
    multiprocessing.get_logger().debug("Everything ran fine")

    if post_function:
        post_function(simulation)

    temp = {}
    temp["simulation"] = simulation
    temp["path"] = path if simulation.is_file else None
    temp["scores"] = []
    temp["error"] = 0.0
    temp["error_count"] = 0.0
    temp["cadetValues"] = cadetValues
    temp["cadetValuesKEQ"] = cadetValuesKEQ

    if individual is not None:
        temp["individual"] = tuple(individual)
    temp["diff"] = []
    temp["sim_time"] = []
    temp["sim_value"] = []
    temp["exp_value"] = []

    for feature in experiment["scores"]:
        featureType = feature["type"]
        featureName = feature["name"]

        if featureType in cache.scores:
            try:
                (
                    scores,
                    sse,
                    sse_count,
                    sim_time,
                    sim_value,
                    exp_value
                ) = cache.scores[featureType].run(
                    temp, target[experiment["name"]][featureName]
                )
            except TypeError:
                return None

            diff = sim_value - exp_value

            temp["scores"].extend(scores)
            temp["error"] += sse
            temp["error_count"] += sse_count
            temp["diff"].extend(diff)
            temp["sim_time"].append(sim_time)
            temp["sim_value"].append(sim_value)
            temp["exp_value"].append(exp_value)

    return temp


def logError(cache, values, error):
    with cache.error_path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        row = list(values)
        row.append(error.returncode)
        row.append(error.stdout)
        row.append(error.stderr)

        writer.writerow(row)


def repeatSimulation(idx):
    "read the original json file and copy it to a subdirectory for each repeat and change where the target data is written"
    settings_file = Path(sys.argv[1])
    with settings_file.open() as json_data:
        settings = jstyleson.load(json_data)

        baseDir = settings.get("baseDir", None)
        if baseDir is not None:
            baseDir = Path(baseDir)
            settings["resultsDir"] = baseDir / settings["resultsDir"]

        resultDir = Path(settings["resultsDir"]) / str(idx)
        resultDir.mkdir(parents=True, exist_ok=True)

        settings["resultsDirOriginal"] = settings["resultsDir"].as_posix()
        settings["resultsDir"] = resultDir.as_posix()

        if baseDir is not None:
            settings["baseDir"] = baseDir.as_posix()

        new_settings_file = resultDir / settings_file.name
        with new_settings_file.open(mode="w") as json_data:
            jstyleson.dump(settings, json_data, indent=4, sort_keys=True)
        return new_settings_file


def setupMCMC(cache):
    "read the original json file and make an mcmc file based on it with new boundaries"
    settings_file = Path(sys.argv[1])
    with settings_file.open() as json_data:
        settings = jstyleson.load(json_data)
        settings["continueMCMC"] = 0

        baseDir = settings.get("baseDir", None)
        if baseDir is not None:
            baseDir = Path(baseDir)
            settings["resultsDir"] = baseDir / settings["resultsDir"]

        resultDirOriginal = Path(settings["resultsDir"])
        resultsOriginal = resultDirOriginal / "result.h5"
        resultDir = resultDirOriginal / "mcmc_refine"
        resultDir.mkdir(parents=True, exist_ok=True)

        settings["resultsDirOriginal"] = resultDirOriginal.as_posix()
        settings["resultsDir"] = resultDir.as_posix()
        settings["PreviousResults"] = resultsOriginal.as_posix()

        if baseDir is not None:
            settings["baseDir"] = baseDir.as_posix()

        if "mcmc_h5" in settings:
            update_json_mcmc(settings)

        settings["searchMethod"] = "MCMC"

        for experiment in settings["experiments"]:
            foundAbsoluteTime = False
            foundAbsoluteHeight = False
            for feature in experiment["scores"]:
                if feature["type"] == "AbsoluteTime":
                    foundAbsoluteTime = True
                if feature["type"] == "AbsoluteHeight":
                    foundAbsoluteHeight = True
            if foundAbsoluteTime is False:
                experiment["scores"].append(
                    {"name": "AbsoluteTime", "type": "AbsoluteTime"}
                )
            if foundAbsoluteHeight is False:
                experiment["scores"].append(
                    {"name": "AbsoluteHeight", "type": "AbsoluteHeight"}
                )

        if "errorModel" in settings:
            individual = getBestIndividual(cache)
            found = createSimulationBestIndividual(individual, cache)
            for experiment in settings["errorModel"]:
                experiment["file_path"] = found[experiment["name"]]

        new_settings_file = resultDir / settings_file.name
        with new_settings_file.open(mode="w") as json_data:
            jstyleson.dump(settings, json_data, indent=4, sort_keys=False)
        return new_settings_file


def update_json_mcmc(settings):
    data = H5()
    data.filename = settings["mcmc_h5"]
    data.load(paths=["/bounds_change/json"], lock=True)
    prior_parameters = jstyleson.loads(str(data.root.bounds_change.json, "ascii"))

    replace = ['transform', 'min', 'max', 'minKA', 'maxKA',
                      'minKEQ', 'maxKEQ', 'factor', 'minNu', 'maxNu',
                      'minSigma', 'maxSigma', 'minVolume', 'maxVolume',
                      'minArea', 'maxArea', 'minLength', 'maxLength']

    if "parameters_mcmc" in settings:
        parameters = settings["parameters_mcmc"]

        parameters_iter = (parameter for parameter in parameters if parameter['transform'] != 'set_value')
        prior_parameters_iter = (parameter for parameter in prior_parameters if parameter['transform'] != 'set_value')

        for new, prior in zip(parameters_iter, prior_parameters_iter):
            ok_location = "location" in new and new["location"].split("/")[-1] == prior["location"].split("/")[-1]
            ok_location_from = "locationFrom" in new and new["locationFrom"].split("/")[-1] == prior["locationFrom"].split("/")[-1]

            if ok_location or ok_location_from:
                #update just the location data everthing else needs to remain the same
                for key, value in new.items():
                    if key in replace:
                        new[key] = prior[key]
            else:
                multiprocessing.get_logger().info(
                    "parameters_mcmc does not have the same transform and variables in the same order as the prior, MCMC cannot continue until this is fixed"
                )
                sys.exit()     
    else:
        parameters = prior_parameters
       
    settings["parameters"].extend(parameters)


def setupAltFeature(cache, name):
    "read the original json file and make an mcmc file based on it with new boundaries"
    settings_file = Path(sys.argv[1])
    with settings_file.open() as json_data:
        settings = jstyleson.load(json_data)

        baseDir = settings.get("baseDir", None)
        if baseDir is not None:
            baseDir = Path(baseDir)
            settings["resultsDir"] = baseDir / settings["resultsDir"]

        resultDirOriginal = Path(settings["resultsDir"])
        resultsOriginal = resultDirOriginal / "result.h5"
        resultDir = resultDirOriginal / "altScore" / name
        resultDir.mkdir(parents=True, exist_ok=True)

        settings["resultsDirOriginal"] = resultDirOriginal.as_posix()
        settings["resultsDir"] = resultDir.as_posix()
        settings["PreviousResults"] = resultsOriginal.as_posix()

        if baseDir is not None:
            settings["baseDir"] = baseDir.as_posix()

        settings["searchMethod"] = "AltScore"

        for experiment in settings["experiments"]:
            for feature in experiment["scoresAlt"]:
                if feature["name"] == name:
                    experiment["scores"] = feature["scores"]
            del experiment["scoresAlt"]

        new_settings_file = resultDir / settings_file.name
        with new_settings_file.open(mode="w") as json_data:
            jstyleson.dump(settings, json_data, indent=4, sort_keys=True)
        return new_settings_file


def createSimulationBestIndividual(individual, cache):
    temp = {}
    for experiment in cache.settings["experiments"]:
        name = experiment["name"]
        templatePath = Path(
            cache.settings["resultsDirMisc"], "template_%s_base.h5" % name
        )
        templateSim = Cadet()
        templateSim.filename = templatePath.as_posix()
        templateSim.load()

        cadetValues, cadetValuesKEQ = set_simulation(
            individual, templateSim, cache.settings, cache, experiment
        )

        bestPath = Path(cache.settings["resultsDirMisc"], "best_%s_base.h5" % name)
        templateSim.filename = bestPath.as_posix()
        templateSim.save()
        temp[name] = bestPath.as_posix()
    return temp


def getBestIndividual(cache):
    "return the path to the best item based on meta min score"
    progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"
    results = H5()
    results.filename = progress_path.as_posix()
    results.load(paths=["/meta_score", "/meta_population"], lock=True)

    idx = numpy.argmin(results.root.meta_score[:, 0])
    individual = results.root.meta_population[idx, :]
    return individual


def copyCSVWithNoise(idx, center, noise):
    "read the original json file and create a new set of simulation data and simulation file in a subdirectory with noise"
    settings_file = Path(sys.argv[1])
    with settings_file.open() as json_data:
        settings = jstyleson.load(json_data)

        baseDir = Path(settings["resultsDir"]) / str(idx)
        baseDir.mkdir(parents=True, exist_ok=True)

        settings["resultsDir"] = str(baseDir)

        del settings["bootstrap"]

        # find CSV files
        for experiment in settings["experiments"]:
            if "csv" in experiment:
                data = numpy.genfromtxt(experiment["csv"], delimiter=",")
                addNoise(data, center, noise)
                csv_path = Path(experiment["csv"])
                new_csv_path = baseDir / csv_path.name
                numpy.savetxt(str(new_csv_path), data, delimiter=",")
                experiment["csv"] = str(new_csv_path)
            for feature in experiment["scores"]:
                if "csv" in feature:
                    data = numpy.genfromtxt(feature["csv"], delimiter=",")
                    addNoise(data, center, noise)
                    csv_path = Path(feature["csv"])
                    new_csv_path = baseDir / csv_path.name
                    numpy.savetxt(str(new_csv_path), data, delimiter=",")
                    feature["csv"] = str(new_csv_path)

        new_settings_file = baseDir / settings_file.name
        with new_settings_file.open(mode="w") as json_data:
            jstyleson.dump(settings, json_data)
        return new_settings_file


def addNoise(array, center, noise):
    "add noise to an array"
    maxValue = numpy.max(array[:, 1])
    randomNoise = numpy.random.normal(center, noise * maxValue, len(array))
    array[:, 1] += randomNoise


def bestMinScore(hof):
    "find the best score based on the minimum of the scores"
    idxMax = numpy.argmax([min(i.fitness.values) for i in hof])
    return hof[idxMax]


def fracStat(time_center, value):
    mean_time = numpy.sum(time_center * value) / numpy.sum(value)
    variance_time = numpy.sum((time_center - mean_time) ** 2 * value) / numpy.sum(value)
    skew_time = numpy.sum((time_center - mean_time) ** 3 * value) / numpy.sum(value)

    mean_value = numpy.sum(time_center * value) / numpy.sum(time_center)
    variance_value = numpy.sum((value - mean_value) ** 2 * time_center) / numpy.sum(
        time_center
    )
    skew_value = numpy.sum((value - mean_value) ** 3 * time_center) / numpy.sum(
        time_center
    )

    return mean_time, variance_time, skew_time, mean_value, variance_value, skew_value


def fractionate_spline(start_seq, stop_seq, spline):
    temp = []
    for (start, stop) in zip(start_seq, stop_seq):
        temp.append(spline.integral(start, stop) / (stop - start))
    return numpy.array(temp)


def find_opt_poly(x, y, index):
    if index == 0:
        indexes = numpy.array([index, index + 1, index + 2])
    elif index == (len(x) - 1):
        indexes = numpy.array([index - 2, index - 1, index])
    else:
        indexes = numpy.array([index - 1, index, index + 1])

    x = x[indexes]
    y = y[indexes]

    if x[0] > x[-1]:  # need to invert order
        x = x[::-1]
        y = y[::-1]

    try:
        poly, res = numpy.polynomial.Polynomial.fit(x, y, 2, full=True)
    except numpy.linalg.LinAlgError:
        multiprocessing.get_logger().error("Polyfit failed for  %s %s", x, y)
    try:
        root = poly.deriv().roots()[0]
    except IndexError:
        # this happens if all y values are the same in which case just take the center value
        root = x[0]
    root = numpy.clip(root, x[0], x[1])
    return root, x, y


def metaCSV(cache):
    repeat = int(cache.settings["repeat"])

    generations = []
    timeToComplete = []
    timePerGeneration = []
    totalCPUTime = []
    paretoFront = []
    avergeScore = []
    minimumScore = []
    productScore = []
    paretoAverageScore = []
    paretoMinimumScore = []
    paretoProductScore = []
    population = None
    dimensionIn = None
    dimensionOut = None
    searchMethod = None

    # read base progress csv and append each score
    base_dir = Path(cache.settings["resultsDirOriginal"])
    paths = [
        base_dir / "progress.csv",
    ]
    for idx in range(repeat):
        paths.append(base_dir / str(idx) / "progress.csv")

    for idx, path in enumerate(paths):
        data = pandas.read_csv(path)

        if idx == 0:
            population = data["Population"].iloc[-1]
            dimensionIn = data["Dimension In"].iloc[-1]
            dimensionOut = data["Dimension Out"].iloc[-1]
            searchMethod = data["Search Method"].iloc[-1]

        generations.append(data["Generation"].iloc[-1])
        timeToComplete.append(data["Elapsed Time"].iloc[-1])
        timePerGeneration.append(numpy.mean(data["Generation Time"]))
        paretoFront.append(data["Pareto Front"].iloc[-1])
        avergeScore.append(data["Average Score"].iloc[-1])
        minimumScore.append(data["Minimum Score"].iloc[-1])
        productScore.append(data["Product Score"].iloc[-1])
        totalCPUTime.append(data["Total CPU Time"].iloc[-1])
        paretoAverageScore = [data["Pareto Mean Average Score"].iloc[-1]]
        paretoMinimumScore = [data["Pareto Mean Minimum Score"].iloc[-1]]
        paretoProductScore = [data["Pareto Mean Product Score"].iloc[-1]]
        paretoProductScore = [data["Pareto Mean Product Score"].iloc[-1]]

    meta_progress = base_dir / "meta_progress.csv"
    with meta_progress.open("w", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)

        writer.writerow(
            [
                "Population",
                "Dimension In",
                "Dimension Out",
                "Search Method",
                "Generation",
                "Generation STDDEV",
                "Elapsed Time",
                "Elapsed Time STDDEV",
                "Generation Time",
                "Geneation Time STDDEV",
                "Pareto Front",
                "Paret Front STDDEV",
                "Average Score",
                "Average Score STDDEV",
                "Minimum Score",
                "Minimum Score STDDEV",
                "Product Score",
                "Product Score STDDEV",
                "Pareto Mean Average Score",
                "Pareto Mean Average Score STDDEV",
                "Pareto Mean Minimum Score",
                "Pareto Mean Minimum Score STDDEV",
                "Pareto Mean Product Score",
                "Pareto Mean Product Score STDDEV",
                "Total CPU Time",
                "Total CPU Time STDDEV",
            ]
        )

        writer.writerow(
            [
                population,
                dimensionIn,
                dimensionOut,
                searchMethod,
                numpy.mean(generations),
                numpy.std(generations),
                numpy.mean(timeToComplete),
                numpy.std(timeToComplete),
                numpy.mean(timePerGeneration),
                numpy.std(timePerGeneration),
                numpy.mean(paretoFront),
                numpy.std(paretoFront),
                numpy.mean(avergeScore),
                numpy.std(avergeScore),
                numpy.mean(minimumScore),
                numpy.std(minimumScore),
                numpy.mean(productScore),
                numpy.std(productScore),
                numpy.mean(paretoAverageScore),
                numpy.std(paretoAverageScore),
                numpy.mean(paretoMinimumScore),
                numpy.std(paretoMinimumScore),
                numpy.mean(paretoProductScore),
                numpy.std(paretoProductScore),
                numpy.mean(totalCPUTime),
                numpy.std(totalCPUTime),
            ]
        )


def update_result_data(cache, ind, fit, result_data, results, meta_scores):
    if result_data is not None and results is not None:
        result_data["input"].append(tuple(ind))
        result_data["output"].append(tuple(fit))
        result_data["output_meta"].append(tuple(meta_scores))

        for result in results.values():
            result_data["input_transform"].append(tuple(result["cadetValues"]))
            result_data["input_transform_extended"].append(
                tuple(result["cadetValuesKEQ"])
            )

            # All results have the same parameter set so we only need the first one
            break

        if cache.fullTrainingData:
            if "results" not in result_data:
                result_data["results"] = {}

            if "times" not in result_data:
                result_data["times"] = {}

            for experimentName, experiment in results.items():
                units_used = cache.target[experimentName]["units_used"]
                sim = experiment["simulation"]
                times = sim.root.output.solution.solution_times

                timeName = "%s_time" % experimentName

                if timeName not in result_data["times"]:
                    result_data["times"][timeName] = times

                for unitName in units_used:
                    unit = sim.root.output.solution[unitName]
                    for solutionName, solution in unit.items():
                        if solutionName.startswith("solution_outlet_comp"):
                            comp = solutionName.replace("solution_outlet_comp_", "")

                            name = "%s_%s_%s" % (experimentName, unitName, comp)

                            if name not in result_data["results"]:
                                result_data["results"][name] = []

                            result_data["results"][name].append(tuple(solution))


def calcFitness(scores_orig, cache):
    scores = numpy.array(scores_orig)[cache.meta_mask]

    if cache.allScoreSSE and cache.MultiObjectiveSSE is False:
        scores = (numpy.sum(scores),)
    return tuple(scores)


def process_population(
    cache,
    population,
    fitnesses,
    writer,
    csvfile,
    halloffame,
    meta_hof,
    progress_hof,
    generation,
    result_data=None,
):
    csv_lines = []
    meta_csv_lines = []

    made_progress = False

    if meta_hof:
        best_min = max([i.fitness.values[2] for i in meta_hof.items])
    else:
        # just in case if the SSE score is used this will make sure that it has to be greater
        best_min = -1e308

    lookup = create_lookup(population)

    last_time = time.time()
    elapsed = cache.progress_elapsed_time

    for idx, result in enumerate(fitnesses):

        if (time.time() - last_time) > elapsed:
            percent = idx / len(population)
            multiprocessing.get_logger().info(
                "Generation %s approximately %.1f %% complete with %s/%s done",
                generation,
                percent * 100,
                idx,
                len(population),
            )
            last_time = time.time()

        fit, csv_line, meta_score, results, individual = result

        ind = pop_lookup(lookup, individual)

        save_name_base = hashlib.md5(
            str(list(ind)).encode("utf-8", "ignore")
        ).hexdigest()

        ind.fitness.values = calcFitness(fit, cache)
        ind.csv_line = [time.ctime(), save_name_base] + csv_line

        ind_meta = pop.Individual(ind)

        ind_meta.fitness.values = meta_score
        ind_meta.csv_line = [time.ctime(), save_name_base] + csv_line

        update_result_data(cache, ind, fit, result_data, results, meta_score)

        if csv_line:
            csv_lines.append([time.ctime(), save_name_base] + csv_line)

            if halloffame is not None:
                onFront, significant = pareto.updateParetoFront(halloffame, ind)
                if onFront and not cache.metaResultsOnly:
                    processResults(save_name_base, ind, cache, results)

            onFrontMeta, significant_meta = pareto.updateParetoFront(meta_hof, ind_meta)
            if onFrontMeta:
                meta_csv_lines.append([time.ctime(), save_name_base] + csv_line)
                processResultsMeta(save_name_base, ind, cache, results)

            # If this is None progress can never be made and the algorithm will terminate, it should only be
            # None if used with algorithms like multistart or scoretest which don't need progress
            if progress_hof is not None:
                onFrontProgress, significant_progress = pareto.updateParetoFront(
                    copy.deepcopy(progress_hof), ind_meta
                )
                if significant_progress:
                    made_progress = True
                    pareto.updateParetoFront(progress_hof, ind_meta)

    writer.writerows(csv_lines)

    new_best_min = max([i.fitness.values[2] for i in meta_hof.items])

    # if the min value is zero then use the old method to determine progress of adding to the pareto front
    # this should catch later stages where real progress is not being made but you have a variable that is not identifiable
    # if new_best_min > 0:
    #    if new_best_min > best_min:
    #        made_progress = True
    #    else:
    #        made_progress = False

    # flush before returning
    csvfile.flush()

    if made_progress:
        if generation != cache.lastProgressGeneration:
            cache.generationsOfProgress += 1
            cache.lastProgressGeneration = generation
    elif generation != cache.lastProgressGeneration:
        cache.generationsOfProgress = 0

    path_meta_csv = cache.settings["resultsDirMeta"] / "results.csv"
    with path_meta_csv.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        writer.writerows(meta_csv_lines)

    cleanupFront(cache, halloffame, meta_hof)
    writeMetaFront(cache, meta_hof, path_meta_csv)

    stalled = (generation - cache.lastProgressGeneration) >= cache.stallGenerations
    stallWarn = (generation - cache.lastProgressGeneration) >= cache.stallCorrect
    progressWarn = cache.generationsOfProgress >= cache.progressCorrect

    return stalled, stallWarn, progressWarn


def eval_population(
    cache,
    invalid_ind,
    writer,
    csvfile,
    halloffame,
    meta_hof,
    progress_hof,
    generation,
    result_data=None,
):
    return eval_population_base(
        cache.eval.evaluate,
        cache,
        invalid_ind,
        writer,
        csvfile,
        halloffame,
        meta_hof,
        progress_hof,
        generation,
        result_data,
    )


def eval_population_final(
    cache,
    invalid_ind,
    writer,
    csvfile,
    halloffame,
    meta_hof,
    progress_hof,
    generation,
    result_data=None,
):
    return eval_population_base(
        cache.eval.evaluate_final,
        cache,
        invalid_ind,
        writer,
        csvfile,
        halloffame,
        meta_hof,
        progress_hof,
        generation,
        result_data,
    )


def eval_population_base(
    evaluate,
    cache,
    invalid_ind,
    writer,
    csvfile,
    halloffame,
    meta_hof,
    progress_hof,
    generation,
    result_data=None,
):
    fitnesses = cache.map_function(evaluate, map(list, invalid_ind))

    return process_population(
        cache,
        invalid_ind,
        fitnesses,
        writer,
        csvfile,
        halloffame,
        meta_hof,
        progress_hof,
        generation,
        result_data,
    )


def get_grad_tolerance(cache, name):
    smallest_peak = cache.target[name]["smallest_peak"]
    largest_peak = cache.target[name]["largest_peak"]

    grad_tol = cache.abstolFactorGrad * smallest_peak
    grad_tol_min = cache.abstolFactorGradMax * largest_peak

    return max([grad_tol, grad_tol_min])


def get_evo_tolerance(cache, name):
    smallest_peak = cache.target[name]["smallest_peak"]
    largest_peak = cache.target[name]["largest_peak"]

    evo_tol = cache.abstolFactor * smallest_peak
    evo_tol_min = cache.abstolFactorGradMax * largest_peak

    return max([evo_tol, evo_tol_min])


def writeMetaFront(cache, meta_hof, path_meta_csv):
    new_data = [individual.csv_line for individual in meta_hof.items]
    new_data = pandas.DataFrame(new_data, columns=cache.headers)

    new_data.to_csv(path_meta_csv, quoting=csv.QUOTE_ALL, index=False)
    new_data.to_excel(cache.settings["resultsDirMeta"] / "results.xlsx", index=False)


def processResultsGrad(save_name_base, individual, cache, results):
    saveExperiments(
        save_name_base,
        cache.settings,
        cache.target,
        results,
        cache.settings["resultsDirGrad"],
        "%s_%s_GRAD.h5",
    )


def processResults(save_name_base, individual, cache, results):
    saveExperiments(
        save_name_base,
        cache.settings,
        cache.target,
        results,
        cache.settings["resultsDirEvo"],
        "%s_%s_EVO.h5",
    )


def processResultsMeta(save_name_base, individual, cache, results):
    saveExperiments(
        save_name_base,
        cache.settings,
        cache.target,
        results,
        cache.settings["resultsDirMeta"],
        "%s_%s_meta.h5",
    )


def cleanupFront(cache, halloffame=None, meta_hof=None, grad_hof=None):
    if halloffame is not None:
        cleanDir(Path(cache.settings["resultsDirEvo"]), halloffame)

    if meta_hof is not None:
        cleanDir(Path(cache.settings["resultsDirMeta"]), meta_hof)

    if grad_hof is not None:
        cleanDir(Path(cache.settings["resultsDirGrad"]), grad_hof)


def cleanDir(dir, hof):
    # find all items in directory
    paths = dir.glob("*.h5")

    # make set of items based on removing everything after _
    exists = {str(path.name).split("_", 1)[0] for path in paths}

    # make set of allowed keys based on hall of hame
    allowed = hof.hashes()

    # remove everything not in hall of fame
    remove = exists - allowed

    for save_name_base in remove:
        for path in dir.glob("%s*" % save_name_base):
            path.unlink()


def finish(cache):
    sub.graph_process(cache, "Last", last=True)


def find_outliers(data, lower_percent=10, upper_percent=90):
    lb, ub = numpy.percentile(data, [lower_percent, upper_percent], 0)
    selected = (data >= lb) & (data <= ub)
    bools = numpy.all(selected, 1)
    return selected, bools


def get_confidence(data, lb=5, ub=95):
    lb, ub = numpy.percentile(data, [lb, ub], 0)
    selected = (data >= lb) & (data <= ub)
    bools = numpy.all(selected, 1)
    return data[bools, :]


def test_eta(eta, xl, xu, size):
    "return delta_q log10 power for each eta to use with distribution"
    x = (xu - xl) / 2.0

    delta_1 = (x - xl) / (xu - xl)
    delta_2 = (xu - x) / (xu - xl)

    rand = numpy.random.rand(size)

    mut_pow = 1.0 / (eta + 1.0)

    xy = numpy.zeros(size)
    val = numpy.zeros(size)
    delta_q = numpy.zeros(size)

    xy[rand < 0.5] = 1.0 - delta_1
    val[rand < 0.5] = 2.0 * rand[rand < 0.5] + (1.0 - 2.0 * rand[rand < 0.5]) * xy[
        rand < 0.5
    ] ** (eta + 1)
    delta_q[rand < 0.5] = val[rand < 0.5] ** mut_pow - 1.0

    xy[rand >= 0.5] = 1.0 - delta_2
    val[rand >= 0.5] = 2.0 * (1.0 - rand[rand >= 0.5]) + 2.0 * (
        rand[rand >= 0.5] - 0.5
    ) * xy[rand >= 0.5] ** (eta + 1)
    delta_q[rand >= 0.5] = 1.0 - val[rand >= 0.5] ** mut_pow

    delta_q = delta_q * (xu - xl)

    return delta_q


def confidence_eta(eta, xl, xu):
    seq = test_eta(eta, xl, xu, 100000)
    confidence = numpy.max(numpy.abs(numpy.percentile(seq, [5, 95])))
    mean = numpy.mean(numpy.abs(seq))

    return mean, confidence


def setupSimulation(sim, times, name, cache):
    "set the user solution times to match the times vector and adjust other sim parameters to required settings"

    try:
        del sim.root.input.solver.user_solution_times
    except KeyError:
        pass

    try:
        del sim.root.output
    except KeyError:
        pass

    sim.root.input.solver.user_solution_times = times
    sim.root.input.solver.sections.section_times[-1] = times[-1]
    sim.root.input["return"].split_components_data = 1

    if cache.dynamicTolerance:
        sim.root.input.solver.time_integrator.abstol = get_evo_tolerance(cache, name)
        sim.root.input.solver.time_integrator.reltol = 0.0

    multiprocessing.get_logger().info(
        "%s abstol=%.3g  reltol=%.3g",
        sim.filename,
        sim.root.input.solver.time_integrator.abstol,
        sim.root.input.solver.time_integrator.reltol,
    )

    experiment_name = sim.root.experiment_name
    if isinstance(experiment_name, bytes):
        experiment_name = experiment_name.decode()
    units_used = cache.target[experiment_name]["units_used"]
    for unit in sim.root.input.model.keys():
        if "unit_" in unit:
            sim.root.input["return"][unit].write_solution_particle = 0
            sim.root.input["return"][unit].write_solution_solid = 0
            sim.root.input["return"][unit].write_solution_column_inlet = 0
            sim.root.input["return"][unit].write_solution_inlet = 0

            sim.root.input["return"][unit].write_sens_bulk = 0
            sim.root.input["return"][unit].write_sens_inlet = 0
            sim.root.input["return"][unit].write_sens_particle = 0
            sim.root.input["return"][unit].write_sens_solid = 0
            sim.root.input["return"][unit].write_sens_flux = 0
            sim.root.input["return"][unit].write_sens_volume = 0

            sim.root.input["return"][unit].write_sensdot_bulk = 0
            sim.root.input["return"][unit].write_sensdot_inlet = 0
            sim.root.input["return"][unit].write_sensdot_outlet = 0
            sim.root.input["return"][unit].write_sensdot_particle = 0
            sim.root.input["return"][unit].write_sensdot_solid = 0
            sim.root.input["return"][unit].write_sensdot_flux = 0
            sim.root.input["return"][unit].write_sensdot_volume = 0

            if unit in units_used:
                sim.root.input["return"][unit].write_solution_column_outlet = 1
                sim.root.input["return"][unit].write_solution_outlet = 1
                sim.root.input["return"][unit].write_sens_outlet = 1
            else:
                sim.root.input["return"][unit].write_solution_column_outlet = 0
                sim.root.input["return"][unit].write_solution_outlet = 0
                sim.root.input["return"][unit].write_sens_outlet = 0
    sim.root.input.solver.nthreads = 1


def biasSimulation(simulation, experiment, cache):

    bias_simulation = Cadet(simulation.root)

    if cache.errorBias:
        name = experiment["name"]

        error_model = None
        for error in cache.settings["errorModel"]:
            if error["name"] == name:
                error_model = error
                break

        delay_settings = error_model["delay"]
        flow_settings = error_model["flow"]
        load_settings = error_model["load"]

        nsec = bias_simulation.root.input.solver.sections.nsec

        delays = numpy.ones(nsec) * sum(delay_settings) / 2.0

        synthetic_error.pump_delay(bias_simulation, delays)

        flow = numpy.ones(nsec) * flow_settings[0]

        synthetic_error.error_flow(bias_simulation, flow)

        load = numpy.ones(nsec) * load_settings[0]

        synthetic_error.error_load(bias_simulation, load)

    return bias_simulation


def setupLog(log_directory, log_name):
    logger = multiprocessing.get_logger()
    logger.propagate = False
    logger.setLevel(logging.INFO)

    logFormatter = logging.Formatter(
        "%(asctime)s %(filename)s %(funcName)s %(lineno)d %(message)s"
    )

    if not logger.handlers:
        stream = logging.StreamHandler(sys.stdout)
        stream.setLevel(logging.INFO)
        stream.setFormatter(logFormatter)

        logger.addHandler(stream)

        sys.stdout = loggerwriter.LoggerWriter(logger.info)
        sys.stderr = loggerwriter.LoggerWriter(logger.warning)

    # create file handler which logs even debug messages
    fh = logging.FileHandler(log_directory / log_name)
    fh.setLevel(logging.INFO)
    fh.setFormatter(logFormatter)

    # add the handlers to the logger
    logger.addHandler(fh)


def getCoreCounts():
    try:
        cpus = int(sys.argv[-1])
    except ValueError:
        # This happens when running in jupyter notebook or if the number can't be found, in either way default to just use availab cpu
        cpus = None
    if cpus:
        return cpus
    else:
        return multiprocessing.cpu_count()


def getMapFunction():
    cores = getCoreCounts()
    if cores == 1:

        multiprocessing.get_logger().info("CADETMatch startup: running single threaded")

        return map
    else:
        if "pool" not in getMapFunction.__dict__:
            pool = multiprocessing.Pool(cores)
            getMapFunction.pool = pool
            multiprocessing.get_logger().info(
                "CADETMatch startup: created a parallel pool of %s workers", cores
            )

        return getMapFunction.pool.imap_unordered


def create_lookup(seq):
    temp = {}
    for i in seq:
        key = tuple(i)
        if key not in temp:
            temp[key] = [
                i,
            ]
        else:
            temp[key].append(i)
    return temp


def pop_lookup(lookup, key):
    temp = lookup[tuple(key)]
    data = temp.pop(0)
    if not temp:
        del lookup[tuple(key)]
    return data


def process_fraction_csv(csv_file):
    data = pandas.read_csv(csv_file)
    rows, cols = data.shape

    data_headers = data.columns.values.tolist()

    start_times = numpy.array(data.iloc[:, 0])
    stop_times = numpy.array(data.iloc[:, 1])
    components = [int(i) for i in data_headers[2:]]
    fractions = data.iloc[:, 2:]

    return start_times, stop_times, components, fractions


def fractionate_sim(start_times, stop_times, components, simulation, unit):
    times = simulation.root.output.solution.solution_times

    fracs = {}
    for component in components:
        sim_value = simulation.root.output.solution[unit][
            "solution_outlet_comp_%03d" % component
        ]
        spline = scipy.interpolate.InterpolatedUnivariateSpline(times, sim_value, ext=1)

        fracs[component] = fractionate_spline(start_times, stop_times, spline)

    return fracs


def get_bins(array, axis=None):
    array = numpy.atleast_2d(array)
    shape = array.shape
    if axis:
        length = shape[axis]
    else:
        length = shape[0]
    h = 2 * scipy.stats.iqr(array, axis=axis) * length**(-1/3)
    bins = (numpy.max(array, axis=axis) - numpy.min(array, axis=axis))/h
    bin = scipy.stats.mstats.gmean(numpy.atleast_1d(bins))
    return round(bin)



import sys

import matplotlib
import matplotlib.style as mplstyle

mplstyle.use("fast")

matplotlib.use("Agg")
import multiprocessing
from pathlib import Path

import matplotlib.pyplot as plt
import numpy
from cadet import H5
from sklearn.preprocessing import MinMaxScaler

from CADETMatch.cache import cache
import CADETMatch.util as util
import CADETMatch.kde_util as kde_util

import seaborn as sns
import pandas


def main():
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    dir_base = Path(cache.settings["resultsDirBase"])
    result_lock = dir_base / "result.lock"

    kde_settings = H5()
    kde_settings.filename = (mcmcDir / "kde_settings.h5").as_posix()
    kde_settings.load(lock=True)

    file = dir_base / "kde_data.h5"

    kde_data = H5()
    kde_data.filename = file.as_posix()
    kde_data.load(lock=True)

    store = kde_settings.root.store

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(str(mcmcDir / "log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], 1 - store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("1 - cross_val_score")
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(str(mcmcDir / "1-log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.savefig(str(mcmcDir / "bandwidth.png"), bbox_inches="tight")
    plt.close()

    times = {}
    values = {}
    for key in kde_data.root.keys():
        if "_time" in key:
            times[key.replace("_time", "")] = kde_data.root[key]
        if "_unit" in key:
            values[key] = kde_data.root[key]

    error_model = mcmcDir / "error_model"
    error_model.mkdir(parents=True, exist_ok=True)

    scaler = MinMaxScaler()
    prob = numpy.exp(kde_settings.root.probability)
    prob = prob[: kde_settings.root.scores.shape[0]]
    prob = numpy.squeeze(scaler.fit_transform(prob.reshape(-1, 1)))

    sort_index = numpy.argsort(prob)

    colors = plt.cm.rainbow(prob)

    for key, value in values.items():
        time = times[key.split("_unit", 1)[0]]
        plt.figure(figsize=[20, 10])
        for idx in sort_index:
            plt.plot(time, value[idx, :], color=colors[idx])
        plt.xlabel("time")
        plt.ylabel("concentration")
        sm = plt.cm.ScalarMappable(
            cmap=plt.cm.rainbow, norm=plt.Normalize(vmin=0, vmax=1)
        )
        plt.colorbar(sm)
        plt.savefig(
            (error_model / ("%s.png" % key)).as_posix(), bbox_inches="tight", dpi=600
        )
        plt.close()

    for exp_name in kde_data.root.errors:
        temp = kde_data.root.errors[exp_name].pump_delays.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10, 10])
            name = "Pump Delays"
            df = pandas.DataFrame(temp, columns=[name,])
            sns.histplot(df[name], bins='auto')
            plt.savefig(
                (error_model / ("%s_pump_delays.png" % exp_name)).as_posix(),
                bbox_inches="tight",
            )
            plt.close()

        temp = kde_data.root.errors[exp_name].flow_rates.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10, 10])
            name = "Flow Rates"
            df = pandas.DataFrame(temp, columns=[name,])
            sns.histplot(df[name], bins='auto')
            plt.savefig(
                (error_model / ("%s_flow_rates.png" % exp_name)).as_posix(),
                bbox_inches="tight",
            )
            plt.close()

        temp = kde_data.root.errors[exp_name].loading_concentrations.reshape(-1, 1)
        if numpy.any(temp > 0):
            plt.figure(figsize=[10, 10])
            name = "Loading Concentration"
            df = pandas.DataFrame(temp, columns=[name,])
            sns.histplot(df[name], bins='auto') 
            plt.savefig(
                (error_model / ("%s_loading_concentrations.png" % exp_name)).as_posix(),
                bbox_inches="tight",
            )
            plt.close()

    (error_model / "scores").mkdir(parents=True, exist_ok=True)
    for idx in range(kde_settings.root.scores.shape[1]):
        plt.figure(figsize=[10, 10])
        temp = kde_settings.root.scores[:, idx]
        name = cache.score_headers[idx]
        df = pandas.DataFrame(temp[temp>0], columns=[name,])
        sns.histplot(df[name], bins='auto') 
        plt.savefig(
            (error_model / "scores" / ("%s.png" % idx)).as_posix(), bbox_inches="tight"
        )
        plt.close()

    (error_model / "scores_mirror_scaled").mkdir(parents=True, exist_ok=True)
    for idx in range(kde_settings.root.scores_mirror_scaled.shape[1]):
        plt.figure(figsize=[10, 10])
        name = cache.score_headers[idx]
        df = pandas.DataFrame(kde_settings.root.scores_mirror_scaled[:, idx], columns=[name,])
        sns.histplot(df[name], bins='auto')   
        plt.savefig(
            (error_model / "scores_mirror_scaled" / ("%s.png" % idx)).as_posix(),
            bbox_inches="tight",
        )
        plt.close()

    (error_model / "scores_mirror").mkdir(parents=True, exist_ok=True)
    for idx in range(kde_settings.root.scores_mirror.shape[1]):
        plt.figure(figsize=[10, 10])
        name = cache.score_headers[idx]
        df = pandas.DataFrame(kde_settings.root.scores_mirror[:, idx], columns=[name,])
        sns.histplot(df[name], bins='auto')   
        plt.savefig(
            (error_model / "scores_mirror" / ("%s.png" % idx)).as_posix(),
            bbox_inches="tight",
        )
        plt.close()


if __name__ == "__main__":
    main()



import array
import csv
import multiprocessing
import time
from pathlib import Path

import CADETMatch.jacobian as jacobian
import CADETMatch.pareto as pareto
import CADETMatch.progress as progress
import CADETMatch.sub as sub
import CADETMatch.util as util
import CADETMatch.pop as pop

name = "ScoreTest"


def run(cache):
    "run the parameter estimation"
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        sim_start = generation_start = time.time()
        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
        }

        population = []
        if "seeds" in cache.settings:
            seed_pop = [
                pop.Individual(
                    [f(v) for f, v in zip(cache.settings["transform"], sublist)]
                )
                for sublist in cache.settings["seeds"]
            ]
            population.extend(seed_pop)

        if cache.metaResultsOnly:
            hof = pareto.DummyFront()
        else:
            hof = pareto.ParetoFront(dimensions=len(cache.WORST),
                similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
            )
        meta_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
            similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
            slice_object=cache.meta_slice
        )
        grad_hof = pareto.ParetoFront(dimensions=len(cache.WORST),
            similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
        )
        progress_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
            similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
            slice_object=cache.meta_slice
        )

        stalled, stallWarn, progressWarn = util.eval_population(
            cache,
            population,
            writer,
            csvfile,
            hof,
            meta_hof,
            progress_hof,
            -1,
            result_data,
        )
        progress.writeProgress(
            cache,
            -1,
            population,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        if cache.settings.get("condTest", None):
            for ind in population:
                J = jacobian.jac(ind, cache)
                multiprocessing.get_logger().info("%s %s", ind, J)

        population = [pop.Individual(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache,
            population,
            writer,
            csvfile,
            hof,
            meta_hof,
            progress_hof,
            0,
            result_data,
        )
        progress.writeProgress(
            cache,
            0,
            population,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        util.finish(cache)
        sub.graph_corner_process(cache, last=True)
        return hof



import numpy

import CADETMatch.util as util
from CADETMatch.abstract.transform import AbstractTransform


class VolumeAreaTransform(AbstractTransform):
    @property
    def name(self):
        return "volume_area"

    @property
    def count(self):
        return 2

    @property
    def count_extended(self):
        return 3

    def transform(self):
        def trans_volume(i):
            return i

        def trans_area(i):
            return i

        return [trans_volume, trans_area]

    grad_transform = transform

    def untransform_inputorder(self, seq):
        volume = seq[0]
        area = seq[1]
        return [volume, area]

    def untransform(self, seq):
        volume, area = self.untransform_inputorder(seq)
        length = volume / area
        values = [area, length]
        headerValues = [area, length, volume]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        temp = self.untransform_matrix_inputorder(matrix)
        values = numpy.zeros(temp.shape)
        values[:, 1] = temp[:, 0] / temp[:, 1]
        values[:, 0] = temp[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        values = numpy.zeros(matrix.shape)
        values[:, 0] = matrix[:, 0]
        values[:, 1] = matrix[:, 1]
        return values

    def setSimulation(self, sim, seq, experiment):
        values, headerValues = self.untransform(seq)

        if (
            self.parameter.get("experiments", None) is None
            or experiment["name"] in self.parameter["experiments"]
        ):
            area_location = self.parameter["area_location"]
            length_location = self.parameter["length_location"]

            sim[area_location.lower()] = values[0]
            sim[length_location.lower()] = values[1]

        return values, headerValues

    def getBounds(self):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minArea = self.parameter["minArea"]
        maxArea = self.parameter["maxArea"]

        minValues = [minVolume, minArea]
        maxValues = [maxVolume, maxArea]

        return minValues, maxValues

    def getGradBounds(self):
        return self.getBounds()

    def getHeaders(self):
        headers = []
        headers.append("Area")
        headers.append("Length")
        headers.append("Volume")
        return headers

    def getHeadersActual(self):
        headers = []
        headers.append("Volume")
        headers.append("Area")
        return headers

    def setBounds(self, parameter, lb, ub):
        parameter["minVolume"] = lb[0]
        parameter["maxVolume"] = ub[0]
        parameter["minArea"] = lb[1]
        parameter["maxArea"] = ub[1]


class NormVolumeAreaTransform(VolumeAreaTransform):
    @property
    def name(self):
        return "norm_volume_area"

    def transform(self):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minArea = self.parameter["minArea"]
        maxArea = self.parameter["maxArea"]

        def trans_volume(i):
            return (i - minVolume) / (maxVolume - minVolume)

        def trans_area(i):
            return (i - minArea) / (maxArea - minArea)

        return [trans_volume, trans_area]

    grad_transform = transform

    def untransform(self, seq):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minArea = self.parameter["minArea"]
        maxArea = self.parameter["maxArea"]

        minValues = numpy.array([minVolume, minArea])
        maxValues = numpy.array([maxVolume, maxArea])

        values = numpy.array(seq)

        values = (maxValues - minValues) * values + minValues

        values = [values[1], values[0] / values[1]]
        headerValues = [values[0], values[1], values[0] * values[1]]
        return values, headerValues

    def grad_untransform(self, seq):
        return self.untransform(seq)[0]

    def untransform_matrix(self, matrix):
        temp = self.untransform_matrix_inputorder(matrix)
        values = numpy.zeros(temp.shape)
        values[:, 0] = temp[:, 1]
        values[:, 1] = temp[:, 0] / temp[:, 1]
        return values

    def untransform_matrix_inputorder(self, matrix):
        minVolume = self.parameter["minVolume"]
        maxVolume = self.parameter["maxVolume"]
        minArea = self.parameter["minArea"]
        maxArea = self.parameter["maxArea"]

        minValues = numpy.array([minVolume, minArea])
        maxValues = numpy.array([maxVolume, maxArea])

        temp = (maxValues - minValues) * matrix + minValues

        values = numpy.zeros(temp.shape)
        values[:, 0] = temp[:, 0]
        values[:, 1] = temp[:, 1]

        return values

    def getBounds(self):
        return [0.0, 0.0], [1.0, 1.0]

    def getGradBounds(self):
        return [self.parameter["min"],], [
            self.parameter["max"],
        ]


plugins = {
    "norm_volume_area": NormVolumeAreaTransform,
    "volume_area": VolumeAreaTransform,
}



import numpy
import scipy.interpolate
import scipy.stats
from addict import Dict

import CADETMatch.score as score
import CADETMatch.smoothing as smoothing
import CADETMatch.util as util

import multiprocessing

# Used to match the front side of a peak (good for breakthrough curves, the drops the derivative lower score)

name = "ShapeFront"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = True
    settings.badScore = 1
    settings.meta_mask = True

    derivative = feature.get("derivative", 1)

    settings.graph = 1
    settings.graph_frac = 0

    if derivative:
        settings.count = 5
        settings.graph_der = 1
    else:
        settings.count = 3
        settings.graph_der = 0

    return settings


def slice_front(times, seq, seq_der, feature):
    "create an array of zeros with only the slice we need in it"
    # resample to 10hz
    new_times = numpy.linspace(times[0], times[-1], int(times[-1] - times[0]) * 10)

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, seq, ext=1)
    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times, seq_der, ext=1)

    seq_resample = spline(new_times)
    seq_der_resample = spline_der(new_times)

    new_seq = numpy.zeros(new_times.shape)
    new_seq_der = numpy.zeros(new_times.shape)

    max_value = numpy.max(seq_resample)

    max_percent = feature.get("max_percent", 0.98)
    min_percent = feature.get("min_percent", 0.02)

    select_max = max_percent * max_value
    select_min = min_percent * max_value

    max_index = numpy.argmax(seq_resample)
    idx_min = numpy.argmin((seq_resample[:max_index] - select_min) ** 2)
    idx_max = numpy.argmin((seq_resample[:max_index] - select_max) ** 2)

    lb_time = new_times[idx_min]
    ub_time = new_times[idx_max]

    min_value = seq_resample[idx_min]
    max_value = seq_resample[idx_max]
    new_seq[idx_min : idx_max + 1] = seq_resample[idx_min : idx_max + 1]
    new_seq_der[idx_min : idx_max + 1] = seq_der_resample[idx_min : idx_max + 1]
    return new_times, new_seq, new_seq_der, min_value, max_value


def slice_front_values(new_times, times, seq, seq_der, min_value, max_value, feature):
    "create an array of zeros with only the slice we need in it"
    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, seq, ext=1)
    spline_der = scipy.interpolate.InterpolatedUnivariateSpline(times, seq_der, ext=1)

    seq_resample = spline(new_times)
    seq_der_resample = spline_der(new_times)

    new_seq = numpy.zeros(new_times.shape)
    new_seq_der = numpy.zeros(new_times.shape)

    max_index = numpy.argmax(seq_resample)

    try:
        idx_min = numpy.argmin((seq_resample[:max_index] - min_value) ** 2)
        idx_max = numpy.argmin((seq_resample[:max_index] - max_value) ** 2)

        new_seq[idx_min : idx_max + 1] = seq_resample[idx_min : idx_max + 1]
        new_seq_der[idx_min : idx_max + 1] = seq_der_resample[idx_min : idx_max + 1]
    except (IndexError, ValueError):
        pass

    return new_seq, new_seq_der


def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(
        sim_data["simulation"], feature
    )
    selected = feature["selected"]

    exp_data_values = feature["value"][selected]
    exp_time_values = feature["time"][selected]
    exp_data_values_der_smooth = feature["exp_data_values_der_smooth"]
    exp_data_values_smooth = feature["exp_data_values_smooth"]
    new_times = feature["new_times"]
    min_value = feature["min_value"]
    max_value = feature["max_value"]

    sim_data_values_smooth, sim_data_values_der_smooth = smoothing.full_smooth(
        exp_time_values,
        sim_data_values,
        feature["critical_frequency"],
        feature["smoothing_factor"],
        feature["critical_frequency_der"],
    )

    ret = slice_front_values(
        new_times,
        exp_time_values,
        sim_data_values_smooth,
        sim_data_values_der_smooth,
        min_value,
        max_value,
        feature,
    )
    sim_data_values_smooth_cut, sim_data_values_der_smooth_cut = ret

    [high, low] = util.find_peak(exp_time_values, sim_data_values_smooth)

    time_high, value_high = high

    pearson, diff_time = score.pearson_spline(
        new_times, sim_data_values_smooth_cut, exp_data_values_smooth
    )

    derivative = feature.get("derivative", 1)

    if derivative:
        pearson_der = score.pearson_offset(
            diff_time,
            new_times,
            sim_data_values_der_smooth_cut,
            exp_data_values_der_smooth,
        )
        [highs_der, lows_der] = util.find_peak(
            exp_time_values, sim_data_values_der_smooth
        )

    temp = [
        pearson,
        feature["value_function"](value_high),
        feature["time_function"](numpy.abs(diff_time)),
    ]
    if derivative:
        temp.extend([pearson_der, feature["value_function_high"](highs_der[1])])

    return (
        temp,
        util.sse(sim_data_values, exp_data_values),
        len(sim_data_values),
        sim_time_values,
        sim_data_values,
        exp_data_values
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    name = "%s_%s" % (sim.root.experiment_name, feature["name"])
    s, crit_fs, crit_fs_der = smoothing.find_smoothing_factors(
        selectedTimes, selectedValues, name, cache
    )
    exp_data_values_smooth, exp_data_values_der_smooth = smoothing.full_smooth(
        selectedTimes, selectedValues, crit_fs, s, crit_fs_der
    )

    [high, low] = util.find_peak(selectedTimes, exp_data_values_der_smooth)

    temp = {}
    temp["peak"] = util.find_peak(selectedTimes, exp_data_values_smooth)[0]

    ret = slice_front(
        selectedTimes, exp_data_values_smooth, exp_data_values_der_smooth, feature
    )
    (
        selectedTimes,
        exp_data_values_smooth,
        exp_data_values_der_smooth,
        min_value,
        max_value,
    ) = ret

    decay = feature.get("decay", 0)

    if decay:
        temp["time_function"] = score.time_function_decay(feature["time"][-1])
    else:
        temp["time_function"] = score.time_function(feature["time"][-1], 0.1 * CV_time)

    temp["value_function"] = score.value_function(temp["peak"][1], abstol)
    temp["value_function_high"] = score.value_function(
        high[1], numpy.abs(high[1]) / 1000
    )
    temp["peak_max"] = max(selectedValues)
    temp["smoothing_factor"] = s
    temp["critical_frequency"] = crit_fs
    temp["critical_frequency_der"] = crit_fs_der
    temp["new_times"] = selectedTimes
    temp["exp_data_values_smooth"] = exp_data_values_smooth
    temp["exp_data_values_der_smooth"] = exp_data_values_der_smooth
    temp["min_value"] = min_value
    temp["max_value"] = max_value
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    derivative = feature.get("derivative", 1)
    temp = ["%s_Similarity" % name, "%s_Value" % name, "%s_Time" % name]

    if derivative:
        temp.extend(["%s_Derivative_Similarity" % name, "%s_Der_High_Value" % name])

    return temp



import subprocess
import sys
from pathlib import Path
import shutil

def clear():
    for path in sorted(Path(sys.argv[1]).resolve().rglob("results")):
        print(path)
        shutil.rmtree(path)

if __name__ == "__main__":
    clear()


import sys

import numpy
import pandas
import scipy.interpolate
from addict import Dict

import CADETMatch.score as score
import CADETMatch.util as util

"""DO NOT USE THIS SCORE. This score only exists for the purpose of a paper to confirm that
this is not an appropriate method to solve fractionation problems.
"""

name = "fractionationSSE"


def get_settings(feature):
    settings = Dict()
    settings.adaptive = False
    settings.badScore = sys.float_info.max
    settings.meta_mask = True
    settings.count = 1
    settings.graph_der = 0
    settings.graph = 0
    settings.graph_frac = 1
    return settings


def run(sim_data, feature):
    "similarity, value, start stop"
    simulation = sim_data["simulation"]
    start = feature["start"]
    stop = feature["stop"]
    comps = feature["comps"]
    data = feature["data"]

    time_center = (start + stop) / 2.0

    times = simulation.root.output.solution.solution_times

    sim_values_sse = []
    exp_values_sse = []

    graph_sim = {}
    graph_exp = {}

    for component in comps:
        exp_values = numpy.array(data[str(component)])
        selected = numpy.isfinite(exp_values)
        sim_value = simulation.root.output.solution[feature["unit"]][
            "solution_outlet_comp_%03d" % component
        ]

        spline = scipy.interpolate.InterpolatedUnivariateSpline(times, sim_value, ext=1)

        fractions = util.fractionate_spline(start[selected], stop[selected], spline)

        exp_values_sse.extend(exp_values[selected])
        sim_values_sse.extend(fractions)

        graph_sim[component] = list(zip(start[selected], stop[selected], fractions))
        graph_exp[component] = list(
            zip(start[selected], stop[selected], exp_values[selected])
        )

    # sort lists
    for key, value in graph_sim.items():
        value.sort()
    for key, value in graph_exp.items():
        value.sort()

    sim_data["graph_exp"] = graph_exp
    sim_data["graph_sim"] = graph_sim

    sse = util.sse(numpy.array(sim_values_sse), numpy.array(exp_values_sse))

    return (
        [
            sse,
        ],
        sse,
        len(sim_values_sse),
        time_center,
        numpy.array(sim_values_sse),
        numpy.array(exp_values_sse)
    )


def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    data = pandas.read_csv(feature["fraction_csv"])

    temp["comps"] = [int(i) for i in data.columns.values.tolist()[2:]]

    temp["data"] = data
    temp["start"] = numpy.array(data.iloc[:, 0])
    temp["stop"] = numpy.array(data.iloc[:, 1])
    temp["unit"] = feature["unit_name"]

    start = numpy.array(data.iloc[:, 0])
    stop = numpy.array(data.iloc[:, 1])

    smallestTime = min(start - stop)
    abstolFraction = abstol * smallestTime

    headers = data.columns.values.tolist()
    temp["peak_max"] = data.iloc[:, 2:].max().min()
    return temp


def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature["name"])
    temp = ["%s_SSE" % name]
    return temp



import sys

import matplotlib
import matplotlib.style as mplstyle

mplstyle.use("fast")

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

import itertools
import logging

# parallelization
import multiprocessing
import os
import warnings
from pathlib import Path

import numpy
import scipy.interpolate
from addict import Dict
from cadet import H5, Cadet
from emcee import autocorr
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.util as util
from CADETMatch.cache import cache


def main(map_function):
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "autocorr.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("autocorr graphing directory %s", os.getcwd())

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    if mcmc_h5.exists():
        resultDir = Path(cache.settings["resultsDirBase"])
        result_lock = resultDir / "result.lock"

        mcmc_store = H5()
        mcmc_store.filename = mcmc_h5.as_posix()
        mcmc_store.load(
            paths=["/full_chain", "/train_full_chain", "/bounds_full_chain"],
            lock=True
        )

        progress_path = Path(cache.settings["resultsDirBase"]) / "result.h5"

        graph_dir = cache.settings["resultsDirSpace"] / "mcmc"

        input_headers = cache.parameter_headers_actual

        for chain in ("full_chain", "train_full_chain", "bounds_full_chain"):
            if chain in mcmc_store.root:
                plot_chain(
                    input_headers, mcmc_store.root[chain], chain, graph_dir / chain
                )


def plot_chain(headers, chain, chain_name, graph_dir):
    graph_dir.mkdir(parents=True, exist_ok=True)

    for i in range(chain.shape[2]):

        fig = figure.Figure(figsize=[15, 7])
        canvas = FigureCanvas(fig)
        graph = fig.add_subplot(1, 1, 1)

        lines = []
        for j in range(chain.shape[0]):
            lines.append(autocorr.function_1d(chain[j, :, i]))
            graph.plot(lines[-1])
        graph.plot(numpy.mean(numpy.array(lines), axis=0), "k", linewidth=4)

        graph.set_xlabel("time")
        graph.set_ylabel("correlation")
        filename = "correlation_%s.png" % (headers[i])
        filename = filename.replace(":", "_").replace("/", "_")

        sum_lines = numpy.sum(numpy.array(lines), axis=0)
        sum_lines = sum_lines / chain.shape[0]
        taus = 2.0 * numpy.cumsum(sum_lines) - 1.0
        tau = taus[autocorr.auto_window(taus, 5)]

        graph.set_title("Correlation graph %s  Tau: %.2f" % (headers[i], tau))
        fig.savefig((graph_dir / filename).as_posix())

        fig = figure.Figure(figsize=[15, 7])
        canvas = FigureCanvas(fig)
        graph = fig.add_subplot(1, 1, 1)
        graph.set_title("Tau graph %s  Tau: %.2f" % (headers[i], tau))
        graph.set_xlabel("time")
        graph.set_ylabel("tau")
        graph.plot(taus, "k", linewidth=4)
        filename = "tau_%s.png" % (headers[i])
        filename = filename.replace(":", "_").replace("/", "_")
        fig.savefig((graph_dir / filename).as_posix())


if __name__ == "__main__":
    map_function = util.getMapFunction()
    main(map_function)
    sys.exit()



import array
import csv
import multiprocessing
import time
from pathlib import Path

import numpy
from cadet import H5

import CADETMatch.jacobian as jacobian
import CADETMatch.pareto as pareto
import CADETMatch.progress as progress
import CADETMatch.sub as sub
import CADETMatch.util as util
import CADETMatch.pop as pop

name = "AltScore"


def run(cache):
    "run the parameter estimation"
    path = Path(cache.settings["resultsDirBase"], cache.settings["csv"])
    with path.open("a", newline="") as csvfile:
        writer = csv.writer(csvfile, delimiter=",", quoting=csv.QUOTE_ALL)
        sim_start = generation_start = time.time()
        result_data = {
            "input": [],
            "output": [],
            "output_meta": [],
            "results": {},
            "times": {},
            "input_transform": [],
            "input_transform_extended": [],
            "strategy": [],
            "mean": [],
            "confidence": [],
        }

        previousResults = cache.settings["PreviousResults"]
        data = H5()
        data.filename = previousResults
        data.load(paths=["/meta_population", "/meta_population_transform"], lock=True)

        meta_population = data.root.meta_population
        meta_population_transform = data.root.meta_population_transform

        population = [pop.Individual(line) for line in meta_population]

        multiprocessing.get_logger().info(
            "altScore starting population %s %s", len(population), population
        )
        multiprocessing.get_logger().info(
            "altScore starting population (transform) %s %s",
            len(meta_population_transform),
            meta_population_transform,
        )

        convert = util.convert_population_inputorder(numpy.array(population), cache)

        multiprocessing.get_logger().info(
            "altScore starting population (convert) %s %s", len(convert), convert
        )

        if cache.metaResultsOnly:
            hof = pareto.DummyFront()
        else:
            hof = pareto.ParetoFront(dimensions=len(cache.WORST),
                similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
            )
        meta_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
            similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
            slice_object=cache.meta_slice
        )
        grad_hof = pareto.ParetoFront(dimensions=len(cache.WORST), 
            similar=pareto.similar, similar_fit=pareto.similar_fit(cache)
        )
        progress_hof = pareto.ParetoFront(dimensions=len(cache.WORST_META),
            similar=pareto.similar, similar_fit=pareto.similar_fit_meta(cache),
            slice_object=cache.meta_slice
        )

        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        stalled, stallWarn, progressWarn = util.eval_population(
            cache,
            invalid_ind,
            writer,
            csvfile,
            hof,
            meta_hof,
            progress_hof,
            -1,
            result_data,
        )
        progress.writeProgress(
            cache,
            -1,
            population,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        if cache.settings.get("condTest", None):
            for ind in invalid_ind:
                J = jacobian.jac(ind, cache)
                multiprocessing.get_logger().info("%s %s", ind, J)

        population = [pop.Individual(i) for i in meta_hof]
        stalled, stallWarn, progressWarn = util.eval_population_final(
            cache,
            population,
            writer,
            csvfile,
            hof,
            meta_hof,
            progress_hof,
            0,
            result_data,
        )
        progress.writeProgress(
            cache,
            0,
            population,
            hof,
            meta_hof,
            grad_hof,
            progress_hof,
            sim_start,
            generation_start,
            result_data,
        )

        util.finish(cache)
        sub.graph_corner_process(cache, last=True)
        return hof



import sys

import matplotlib
import matplotlib.style as mplstyle

mplstyle.use("fast")

matplotlib.use("Agg")

size = 20

matplotlib.rc("font", size=size)  # controls default text sizes
matplotlib.rc("axes", titlesize=size)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=size)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=size)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=size)  # legend fontsize
matplotlib.rc("figure", titlesize=size)  # fontsize of the figure title
matplotlib.rc("figure", autolayout=True)

import itertools
import logging

# parallelization
import multiprocessing
import os
import warnings
from pathlib import Path

import corner
import numpy
import pandas
import scipy.interpolate
import seaborn as sns
from addict import Dict
from cadet import H5
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from mpl_toolkits.mplot3d import Axes3D

import CADETMatch.loggerwriter as loggerwriter
import CADETMatch.util as util
from CADETMatch.cache import cache


def new_range(flat_chain):
    lb_data, mid_data, ub_data = numpy.percentile(flat_chain, [5, 50, 95], 0)

    distance = numpy.max(
        numpy.abs(numpy.array([lb_data - mid_data, ub_data - mid_data])), axis=0
    )

    lb_range = mid_data - 2 * distance
    ub_range = mid_data + 2 * distance

    return numpy.array([lb_range, ub_range])


def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "graph_corner.log")
    cache.setup(sys.argv[1])

    multiprocessing.get_logger().info("graphing directory %s", os.getcwd())

    resultDir = Path(cache.settings["resultsDirBase"])
    result_lock = resultDir / "result.lock"


    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"
    result_h5 = resultDir / "result.h5"

    if mcmc_h5.exists():
        data_mcmc = H5()
        data_mcmc.filename = mcmc_h5.as_posix()
        data_mcmc.load(lock=True)

        data_results = None
    else:
        data_results = H5()
        data_results.filename = result_h5.as_posix()
        data_results.load(lock=True)

        data_mcmc = None

    graphCorner(cache, data_mcmc, data_results)


def plotChain(flat_chain, flat_chain_transform, headers, out_dir, prefix):
    multiprocessing.get_logger().info("plotting chain")
    if len(flat_chain) > 1e5:
        indexes = numpy.random.choice(flat_chain.shape[0], int(1e5), replace=False)
        chain = flat_chain[indexes]
        chain_transform = flat_chain_transform[indexes]
    else:
        chain = flat_chain
        chain_transform = flat_chain_transform

    if len(chain) == 0:
        return

    # stat corner plot
    fig_size = 6 * len(headers)
    fig = corner.corner(
        chain,
        quantiles=(0.16, 0.5, 0.84),
        show_titles=True,
        labels=headers,
        bins=20,
        range=new_range(chain).T,
        use_math_text=True,
        title_fmt=".2g",
    )
    fig.set_size_inches((fig_size, fig_size))
    fig.savefig(str(out_dir / ("%s_corner.png" % prefix)))

    fig = corner.corner(
        chain_transform,
        quantiles=(0.16, 0.5, 0.84),
        show_titles=True,
        labels=headers,
        bins=20,
        range=new_range(chain_transform).T,
        use_math_text=True,
        title_fmt=".2g",
    )
    fig.set_size_inches((fig_size, fig_size))
    fig.savefig(str(out_dir / ("%s_corner_transform.png" % prefix)))


def plotMCMCParam(out_dir, param, chain, header):
    multiprocessing.get_logger().info("plotting mcmc param %s", param)
    shape = chain.shape
    x = numpy.linspace(0, shape[1], shape[1])

    fig = figure.Figure(figsize=[10, 10])
    canvas = FigureCanvas(fig)
    graph = fig.add_subplot(1, 1, 1)

    graph.plot(x, chain[0, :, param], "r", label="5")
    graph.plot(x, chain[1, :, param], "k", label="mean")
    graph.plot(x, chain[2, :, param], "r", label="95")
    graph.fill_between(
        x, chain[0, :, param], chain[2, :, param], facecolor="r", alpha=0.5
    )
    graph.legend()
    graph.set_title(header)
    fig.set_size_inches((12, 12))
    print(header, header.replace("/", "_"))
    fig.savefig(str(out_dir / ("%s.png" % header.replace("/", "_"))))


def plotMCMCVars(out_dir, headers, data):
    for param_idx, header in enumerate(headers):
        if "train_chain_stat" in data.root:
            plotMCMCParam(
                out_dir,
                param_idx,
                data.root.train_chain_stat,
                "Train " + header + str(param_idx),
            )
            plotMCMCParam(
                out_dir,
                param_idx,
                data.root.train_chain_stat_transform,
                "Train " + header + " Transform" + str(param_idx),
            )

        if "run_chain_stat" in data.root:
            plotMCMCParam(
                out_dir,
                param_idx,
                data.root.run_chain_stat,
                "Run " + header + str(param_idx),
            )
            plotMCMCParam(
                out_dir,
                param_idx,
                data.root.run_chain_stat_transform,
                "Run " + header + " Transform" + str(param_idx),
            )


def keep_header(char):
    allowed = set("-0123456789")
    return char in allowed


def clean_header(header):
    splits = header.split()
    temp = [splits[0]]
    for string in splits[1:]:
        temp.append("".join([i for i in string if keep_header(i)]))
    return " ".join(temp)


def graphCorner(cache, data_mcmc=None, data_results=None):
    multiprocessing.get_logger().info("plotting corner plots")
    headers = list(cache.parameter_headers_actual)
    headers = [clean_header(header) for header in headers]

    if data_mcmc is not None:
        data = data_mcmc

        out_dir = cache.settings["resultsDirProgress"]

        if "tau_percent" in data.root:
            multiprocessing.get_logger().info("plotting tau plots")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            tau_percent = numpy.squeeze(data.root.tau_percent)
            max_values = numpy.ones(tau_percent.shape)
            tau_percent = numpy.min([tau_percent, max_values], 0)

            graph.bar(list(range(len(headers))), tau_percent, tick_label=headers)
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "tau_percent.png"))

        if "flat_chain" in data.root:
            multiprocessing.get_logger().info("plot flat chain")
            plotChain(
                data.root.flat_chain,
                data.root.flat_chain_transform,
                headers,
                out_dir,
                "full",
            )

        if "integrated_autocorrelation_time" in data.root:
            iat = data.root.integrated_autocorrelation_time

            multiprocessing.get_logger().info("plot integrated autocorrelation time")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)

            for i in range(1, iat.shape[1]):
                graph.plot(iat[:, 0], iat[:, i], label=i)
            graph.set_title("Integrated Autocorrelation Time")
            graph.set_xlabel("Step")
            graph.set_ylabel("IAT")
            graph.legend()
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time.png"))

            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            tau = cache.MCMCTauMult
            for i in range(1, iat.shape[1]):
                progress = (iat[:, 0] / tau) / iat[:, i]
                progress[progress > 1] = 1
                graph.plot(iat[:, 0], progress, label=i)
            graph.set_title("Integrated Autocorrelation Time Progress")
            graph.set_xlabel("Step")
            graph.set_ylabel("Progress")
            graph.legend()
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "integrated_autocorrelation_time_progress.png"))

        if "mcmc_acceptance" in data.root:
            multiprocessing.get_logger().info("plot mcmc acceptance")
            fig = figure.Figure(figsize=[10, 10])
            canvas = FigureCanvas(fig)
            graph = fig.add_subplot(1, 1, 1)
            graph.plot(util.get_confidence(data.root.mcmc_acceptance))
            graph.set_title("MCMC Acceptance")
            graph.set_xlabel("Step")
            graph.set_ylabel("Acceptance")
            fig.set_size_inches((12, 12))
            fig.savefig(str(out_dir / "mcmc_acceptance.png"))

        plotChain(
            data.root.train_flat_chain,
            data.root.train_flat_chain_transform,
            headers,
            out_dir,
            "train",
        )
        plotMCMCVars(out_dir, headers, data)

    else:
        data = data_results

        if len(data.root.input) > 1e5:
            indexes = numpy.random.choice(
                data.root.input.shape[0], int(1e5), replace=False
            )
            data_input = data.root.input[indexes]
            data_input_transform = data.root.input_transform[indexes]
            weight_min = data.root.output_meta[indexes, 1]
            weight_prod = data.root.output_meta[indexes, 2]
            all_scores = data.root.output[indexes]
        else:
            data_input = data.root.input
            data_input_transform = data.root.input_transform
            weight_min = data.root.output_meta[:, 1]
            weight_prod = data.root.output_meta[:, 2]
            all_scores = data.root.output

        max_scores = numpy.max(all_scores, 1)
        acceptable = max_scores > 0.01

        if numpy.any(acceptable):
            multiprocessing.get_logger().info(
                "graphing remove %s points", len(max_scores) - numpy.sum(acceptable)
            )

            data_input = data_input[acceptable]
            data_input_transform = data_input_transform[acceptable]
            weight_min = weight_min[acceptable]
            weight_prod = weight_prod[acceptable]

        out_dir = cache.settings["resultsDirProgress"]

        create_corner(out_dir, "corner.png", headers, data_input, weights=None)
        create_corner(
            out_dir, "corner_min.png", headers, data_input, weights=weight_min
        )
        create_corner(
            out_dir, "corner_prod.png", headers, data_input, weights=weight_prod
        )

        # transformed entries
        create_corner(
            out_dir, "corner_transform.png", headers, data_input_transform, weights=None
        )
        create_corner(
            out_dir,
            "corner_min_transform.png",
            headers,
            data_input_transform,
            weights=weight_min,
        )
        create_corner(
            out_dir,
            "corner_prod_transform.png",
            headers,
            data_input_transform,
            weights=weight_prod,
        )


def create_corner(dir, filename, headers, data, weights=None):
    if len(data) <= len(headers):
        multiprocessing.get_logger().info(
            "There is not enough data to generate corner plots"
        )
        return
    if numpy.all(numpy.min(data, 0) < numpy.max(data, 0)):
        if weights is None or numpy.max(weights) > numpy.min(weights):
            bounds = new_range(data).T

            fig_size = 6 * len(headers)
            fig = corner.corner(
                data,
                quantiles=(0.16, 0.5, 0.84),
                show_titles=True,
                labels=headers,
                bins=20,
                range=new_range(data).T,
                use_math_text=True,
                title_fmt=".2g",
            )
            fig.set_size_inches((fig_size, fig_size))
            fig.savefig(str(dir / filename))

            df = pandas.DataFrame(data, columns=headers)
            fig = sns.PairGrid(df, diag_sharey=False, height=fig_size)
            fig.map_lower(sns.kdeplot)
            fig.map_upper(sns.scatterplot)
            fig.map_diag(sns.kdeplot, lw=3)

            for row_num, row in enumerate(bounds):
                for col_num, col in enumerate(bounds):
                    if row_num == col_num:
                        fig.axes[row_num, col_num].set_xlim(*row)
                    else:
                        fig.axes[row_num, col_num].set_xlim(*col)
                        fig.axes[row_num, col_num].set_ylim(*row)

            fig.savefig(str(dir / ("sns_%s" % filename)))


if __name__ == "__main__":
    main()



import numpy
import array
import attr
from typing import Iterable

def seq2array(x: Iterable[float]) -> array.array:
    return array.array("d", x)

@attr.s
class Fitness:
    "designed for DEAP compatibility but only minimization, clean and array.array"
    _values = attr.ib(converter=seq2array)
    
    @property
    def wvalues(self):
        "wvalues assumes maximization for some reason to flip the sign"
        return tuple([-1*i for i in  self._values])

    @property
    def values(self):
        return self._values
    
    @values.setter
    def values(self, value):
        self._values = seq2array(value)

    @values.deleter
    def values(self):
        self._values = ()
    
    @property
    def valid(self):
        return len(self._values) != 0
    
    def __array__(self, dtype=None):
        return numpy.frombuffer(self._values, dtype=dtype)
    
    def dominates(self, other, obj=slice(None)):
        """Return true if each objective of *self* is not strictly worse than
        the corresponding objective of *other* and at least one objective is
        strictly better.
        :param obj: Slice indicating on which objectives the domination is
                    tested. The default value is `slice(None)`, representing
                    every objectives.
        """
        dominate = False
        for self_value, other_value in zip(self.values[obj], other.values[obj]):
            if self_value < other_value:
                dominate = True
            elif self_value > other_value:
                return False
        return dominate
    
@attr.s
class Individual:
    value = attr.ib(converter=seq2array)
    fitness = attr.ib(converter=Fitness, default=attr.Factory(list))
    best = attr.ib(default=None)
    csv_line = attr.ib(default='')
    
    @property
    def valid(self):
        return self.fitness.valid
    
    def __array__(self, dtype=None):
        return numpy.frombuffer(self.value, dtype=dtype)

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def __getitem__(self, index):
        return self.value[index]

    def __setitem__(self, index, value):
        self.value[index] = value


import multiprocessing
import sys
import warnings
from pathlib import Path

import joblib
import matplotlib
import matplotlib.style as mplstyle
import numpy
import pandas
import scipy
from addict import Dict
from cadet import H5, Cadet
from sklearn import preprocessing
from sklearn.neighbors import KernelDensity

import CADETMatch.evo as evo
import CADETMatch.kde_util as kde_util
import CADETMatch.smoothing as smoothing
import CADETMatch.util as util
from CADETMatch.cache import cache

mplstyle.use("fast")

matplotlib.use("Agg")

import matplotlib.cm
import matplotlib.pyplot as plt
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

cm_plot = matplotlib.cm.gist_rainbow

import itertools
import logging

import CADETMatch.loggerwriter as loggerwriter

import arviz as az

def get_color(idx, max_colors, cmap):
    return cmap(1.0 * float(idx) / max_colors)


saltIsotherms = {
    b"STERIC_MASS_ACTION",
    b"SELF_ASSOCIATION",
    b"MULTISTATE_STERIC_MASS_ACTION",
    b"SIMPLE_MULTISTATE_STERIC_MASS_ACTION",
    b"BI_STERIC_MASS_ACTION",
}

size = 20

plt.rc("font", size=size)  # controls default text sizes
plt.rc("axes", titlesize=size)  # fontsize of the axes title
plt.rc("axes", labelsize=size)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=size)  # fontsize of the tick labels
plt.rc("ytick", labelsize=size)  # fontsize of the tick labels
plt.rc("legend", fontsize=size)  # legend fontsize
plt.rc("figure", titlesize=size)  # fontsize of the figure title
plt.rc("figure", autolayout=True)

class ArvizSampler:
    def __init__(self, chain, prob):
        self.chain = chain.swapaxes(0, 1)
        self.prob = prob.swapaxes(0, 1)
    
    def get_chain(self):
        return self.chain

    def get_log_prob(self):
        return self.prob

def flatten(chain):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    return flat_chain

def reduce_data(data_chain, probability, size):
    sampler = ArvizSampler(data_chain, probability)
    emcee_data = az.from_emcee(sampler)
    hdi = az.hdi(emcee_data, hdi_prob=0.99).to_array().values

    multiprocessing.get_logger().info("hdi %s", hdi)

    data= flatten(data_chain)

    selected = (data > hdi[:,0]) & (data < hdi[:,1])
    selected = numpy.all(selected, axis=1)

    data = data[selected]

    scaler = preprocessing.RobustScaler().fit(data)

    data = scaler.transform(data)

    if size < data.shape[0]:
        indexes = numpy.random.choice(data.shape[0], size, replace=False)
        data_reduced = data[indexes]
    else:
        data_reduced = data

    return data_reduced, scaler


def get_prior(data_chain, probability, headers):
    multiprocessing.get_logger().info("setting up scaler and reducing data")
    data_reduced, scaler = reduce_data(data_chain, probability, 30000)

    multiprocessing.get_logger().info("finished setting up scaler and reducing data")
    multiprocessing.get_logger().info("data_reduced shape %s", data_reduced.shape)

    kde_ga = KernelDensity(kernel="gaussian")

    kde_ga, bandwidth, store = kde_util.get_bandwidth(kde_ga, data_reduced)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])

    plot_bandwidth(store, mcmcDir)

    multiprocessing.get_logger().info("mle bandwidth: %.2g", bandwidth)

    multiprocessing.get_logger().info("fitting kde with mle bandwidth")

    kde_ga.fit(data_reduced)

    multiprocessing.get_logger().info("finished fitting")

    return kde_ga, scaler

def plot_bandwidth(store, mcmcDir):
    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(str(mcmcDir / "prior_log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], 1 - store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("1 - cross_val_score")
    plt.yscale("log")
    plt.xscale("log")
    plt.savefig(str(mcmcDir / "prior_1-log_bandwidth.png"), bbox_inches="tight")
    plt.close()

    plt.figure(figsize=[10, 10])
    plt.scatter(store[:, 0], store[:, 1])
    plt.xlabel("bandwidth")
    plt.ylabel("cross_val_score")
    plt.savefig(str(mcmcDir / "prior_bandwidth.png"), bbox_inches="tight")
    plt.close()


def process_mle(chain, probability, cache):
    kde, scaler = get_prior(chain, probability, cache.parameter_headers_actual)

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    joblib.dump(scaler, mcmcDir / "kde_prior_scaler.joblib")
    joblib.dump(kde, mcmcDir / "kde_prior.joblib")


def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings["resultsDirLog"], "prior.log")
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings["resultsDirMCMC"])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    resultDir = Path(cache.settings["resultsDirBase"])
    result_lock = resultDir / "result.lock"

    mcmc_store = H5()
    mcmc_store.filename = mcmc_h5.as_posix()
    mcmc_store.load(paths=["/full_chain", "/mcmc_acceptance", "/run_probability"], lock=True)

    process_mle(mcmc_store.root.full_chain, 
                mcmc_store.root.run_probability,
                cache)


if __name__ == "__main__":
    main()




import copy
from operator import eq
import hashlib
import CADETMatch.pop as pop
import bisect
import itertools

import numpy
class ParetoFront:
    "Pareto front designed to be compatible with DEAP"

    def __init__(self, dimensions, similar=None, similar_fit=None, slice_object=slice(None)):
        if similar is None:
            similar = eq
        
        if similar_fit is not None:
            self.similar_fit = similar_fit
        else:
            self.similar_fit = eq
        
        self.slice_object = slice_object
        self.keys = list()
        self.items = list()
        self.similar = similar

        #due to the crowding measures if solutions are very close together it is possible for the very
        #best solution for one of the goals to not be stored so they should be kept separately and
        #merged in
        self.dimensions = dimensions
        self.best_keys = [1e308]*dimensions
        self.best_items = [None]*dimensions

    def insert(self, item):
        """Insert a new item in sorted order"""
        item = copy.deepcopy(item)
        i = bisect.bisect(self.keys, item.fitness)
        self.items.insert(i, item)
        self.keys.insert(i, item.fitness)

    def remove(self, index):
        del self.keys[index]
        del self.items[index]

    def clear(self):
        del self.items[:]
        del self.keys[:]

    def __len__(self):
        return len(self.items)

    def __getitem__(self, i):
        return self.items[i]

    def __iter__(self):
        return iter(self.items)

    def __reversed__(self):
        return reversed(self.items)

    def update(self, population):
        """Update the Pareto front hall of fame with the *population* by adding
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.

        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        new_members = []
        significant = []
        slice_object = self.slice_object
        pareto_length = len(self)

        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []
            
            for idx in range(len(ind.fitness.values)):
                if ind.fitness.values[idx] < self.best_keys[idx]:
                    self.best_keys[idx] = ind.fitness.values[idx]
                    self.best_items[idx] = copy.deepcopy(ind)

            for i, hofer in enumerate(self):  # hofer = hall of famer
                if not dominates_one and hofer.fitness.dominates(
                    ind.fitness, obj=slice_object
                ):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hofer.fitness, obj=slice_object):
                    dominates_one = True
                    to_remove.append(i)
                    significant.append(
                        not self.similar_fit(ind.fitness.values, hofer.fitness.values)
                    )
                elif self.similar_fit(
                    ind.fitness.values, hofer.fitness.values
                ) and self.similar(ind.value, hofer.value):
                    has_twin = True
                    break

            for i in reversed(to_remove):  # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                # if the pareto front is empty or a new item is added to the pareto front that is progress
                # however if there is only a single objective and it does not significantly dominate then
                # don't count that as significant progress
                if pareto_length == 0:
                    significant.append(True)
                elif self.dimensions > 1:
                    if len(significant) == 0 or (len(significant) and any(significant)):
                        significant.append(True)

                self.insert(ind)
                new_members.append(ind)

        return new_members, any(significant)

    def hashes(self):
        hashes = {
            hashlib.md5(str(list(individual.value)).encode("utf-8", "ignore")).hexdigest()
            for individual in itertools.chain(self.items, self.best_items) if individual is not None
        }
        return hashes 


    def totalEntries(self):
        population = []

        for ind in itertools.chain(self.items, self.best_items):
            if ind is not None and ind.value not in population:
                population.append(ind.value)

        return len(population)


    def getEntries(self):
        population = []
        fitnesses  = []

        for ind in itertools.chain(self.items, self.best_items):
            if ind is not None and ind.value not in population:
                population.append(ind.value)
                fitnesses.append(ind.fitness.values)

        return numpy.array(population), numpy.array(fitnesses)



    def getBestScores(self):
        items = [i.fitness.values for i in itertools.chain(self.items, self.best_items) if i is not None]
        data = numpy.array(items)
        return numpy.min(data, 0)

class DummyFront(ParetoFront):
    "Pareto front designed to be compatible with DEAP"

    def __init__(self, dimensions=None, similar=None):
        "This is here for API compatibility, don't do anything"
        self.items = []
        self.best_items = []
        return None

    def update(self, population):
        "do not put anything in this front, it is just needed to maintain compatibility"
        return [], False


def similar(a, b):
    "for minimization the rtol needs to be fairly high otherwise the pareto front contains too many entries"
    a = numpy.frombuffer(a)
    b = numpy.frombuffer(b)
    return numpy.allclose(a,b, rtol=1e-1)


def similar_fit_meta_split(a, b):
    a = numpy.frombuffer(a)
    b = numpy.frombuffer(b)
    return numpy.allclose(a[:3],b[:3], rtol=1e-1)


def similar_fit_meta_sse(a, b):
    "SSE is negative and in the last slot and the only score needed"
    a = numpy.frombuffer(a)
    b = numpy.frombuffer(b)
    return numpy.allclose(a[-2],b[-2], rtol=1e-1)


def similar_fit_meta(cache):
    if cache.allScoreSSE and not cache.MultiObjectiveSSE:
        return similar_fit_meta_sse
    else:
        return similar_fit_meta_split


def similar_fit(cache):
    return similar


def updateParetoFront(halloffame, offspring):
    new_members, significant = halloffame.update(
        [
            offspring,
        ]
    )
    return bool(new_members), significant